# https://altair-viz.github.io/gallery/world_map.html

import altair as alt
from vega_datasets import data

# Data generators for the background
sphere = alt.sphere()
graticule = alt.graticule()

# Source of land data
source = alt.topo_feature(data.world_110m.url, 'countries')

# Layering and configuring the components
alt.layer(
    alt.Chart(sphere).mark_geoshape(fill='lightblue'),
    alt.Chart(graticule).mark_geoshape(stroke='white', strokeWidth=0.5),
    alt.Chart(source).mark_geoshape(fill='ForestGreen', stroke='black')
).project(
    'naturalEarth1'
).properties(width=600, height=400).configure_view(stroke=None)



# https://altair-viz.github.io/gallery/layered_bar_chart.html
# With explicit temporal encoding
import altair as alt
from vega_datasets import data

source = data.iowa_electricity()

alt.Chart(source).mark_bar(opacity=0.7).encode(
    x='year:T',
    y=alt.Y('net_generation:Q', stack=None),
    color="source",
)



import re
import warnings

from . import SqlConnection, CsvReadOptions

from typing import Dict, Optional
from distutils.version import LooseVersion

import duckdb
import pyarrow as pa
import pyarrow.feather
import pandas as pd
import logging
import uuid

# Table suffix name to use for raw registered table
RAW_PREFIX = "_vf_raw_"


def duckdb_type_name_to_pyarrow_type(duckdb_type: str) -> pa.DataType:
    duckdb_type = str(duckdb_type).upper()
    if duckdb_type in ("VARCHAR", "JSON", "CHAR", "CATEGORICAL"):
        return pa.string()
    elif duckdb_type in ("REAL", "FLOAT4", "FLOAT"):
        return pa.float32()
    elif duckdb_type in ("DOUBLE", "FLOAT8"):
        return pa.float64()
    elif duckdb_type in ("TINYINT", "INT1"):
        return pa.int8()
    elif duckdb_type in ("SMALLINT", "INT2", "SHORT"):
        return pa.int16()
    elif duckdb_type in ("INTEGER", "INT4", "INT", "SIGNED"):
        return pa.int32()
    elif duckdb_type in ("BIGINT", "INT8", "LONG"):
        return pa.int64()
    elif duckdb_type.startswith("DECIMAL"):
        matches = re.findall(r"\d+", duckdb_type)
        precision = int(matches[0])
        scale = int(matches[1])
        return pa.decimal128(precision, scale)
    elif duckdb_type == "UTINYINT":
        return pa.uint8()
    elif duckdb_type == "USMALLINT":
        return pa.uint16()
    elif duckdb_type == "UINTEGER":
        return pa.uint32()
    elif duckdb_type == "UBIGINT":
        return pa.uint64()
    elif duckdb_type == "BOOLEAN":
        return pa.bool_()
    elif duckdb_type == "DATE":
        return pa.date32()
    elif duckdb_type in ("TIMESTAMP", "TIMESTAMP_MS"):
        return pa.timestamp("ms")
    elif duckdb_type == "TIMESTAMP_NS":
        return pa.timestamp("ns")
    elif duckdb_type == "TIMESTAMP WITH TIME ZONE":
        return pa.timestamp("ms", tz="UTC")
    else:
        raise ValueError(f"Unexpected DuckDB type: {duckdb_type}")


def duckdb_relation_to_schema(rel: duckdb.DuckDBPyRelation) -> pa.Schema:
    schema_fields = {}
    for col, type_name in zip(rel.columns, rel.dtypes):
        try:
            type_ = duckdb_type_name_to_pyarrow_type(type_name)
            schema_fields[col] = type_
        except ValueError:
            # Skip columns with unrecognized types
            pass
    return pa.schema(schema_fields)


def pyarrow_type_to_duckdb_type_name(field_type: pa.Schema) -> Optional[str]:
    if field_type in (pa.utf8(), pa.large_utf8()):
        return "VARCHAR"
    elif field_type in (pa.float16(), pa.float32()):
        return "FLOAT"
    elif field_type == pa.float64():
        return "DOUBLE"
    elif field_type == pa.int8():
        return "TINYINT"
    elif field_type == pa.int16():
        return "SMALLINT"
    elif field_type == pa.int32():
        return "INTEGER"
    elif field_type == pa.int64():
        return "BIGINT"
    elif field_type == pa.uint8():
        return "UTINYINT"
    elif field_type == pa.uint16():
        return "USMALLINT"
    elif field_type == pa.uint32():
        return "UINTEGER"
    elif field_type == pa.uint64():
        return "UBIGINT"
    elif field_type == pa.bool_():
        return "BOOLEAN"
    elif field_type == pa.date32():
        return "DATE"
    else:
        return None


def pyarrow_schema_to_select_replace(schema: pa.Schema, table_name: str) -> str:
    """
    Build `SELECT * REPLACE(...) from table_name` query that casts columns
    to match the provided pyarrow schema.

    This is needed because sometimes the resulting DuckDB column types won't exactly
    match those that DataFusion expects (e.g. DuckDB returning a DECIMAL(5) instead of
    and int64). Types that are not covered by `pyarrow_type_to_duckdb_type_name` above
    are passed through as-is.
    """
    replaces = []
    for field_index in range(len(schema)):
        field = schema.field(field_index)
        field_name = field.name
        field_type = field.type

        quoted_column = quote_column(field_name)
        duckdb_type = pyarrow_type_to_duckdb_type_name(field_type)
        if duckdb_type:
            replaces.append(f"{quoted_column}::{duckdb_type} as {quoted_column}")

    if replaces:
        replace_csv = ", ".join(replaces)
        return f"SELECT * REPLACE({replace_csv}) FROM {table_name}"
    else:
        return f"SELECT * FROM {table_name}"


class DuckDbConnection(SqlConnection):
    def __init__(self, connection: duckdb.DuckDBPyConnection = None, fallback: bool = True, verbose: bool = False):
        # Validate duckdb version
        if LooseVersion(duckdb.__version__) < LooseVersion("0.7.0"):
            raise ImportError(
                f"The VegaFusion DuckDB connection requires at least DuckDB version 0.7.0\n"
                f"Found version {duckdb.__version__}"
            )

        self._fallback = fallback
        self._verbose = verbose
        self._temp_tables = set()

        if connection is None:
            connection = duckdb.connect()

            # Install and load the httpfs extension only if we are creating the duckdb connection
            # here. If a connection was passed in, don't assume it has internet access and the
            # ability to install extensions
            try:
                connection.install_extension("httpfs")
                connection.load_extension("httpfs")
            except (IOError, duckdb.IOException, duckdb.InvalidInputException) as e:
                warnings.warn(f"Failed to install and load the DuckDB httpfs extension:\n{e}")

            # Use a less round number for pandas_analyze_sample (default is 1000)
            connection.execute("SET GLOBAL pandas_analyze_sample=1007")

        # The icu extension is pre-bundled in Python, so no need to install it
        connection.load_extension("icu")

        self.conn = connection
        self.logger = logging.getLogger("DuckDbConnection")

        self._registered_table_schemas = dict()
        # Call self.tables to warm the cache of table schemas
        self.tables()

    @classmethod
    def dialect(cls) -> str:
        return "duckdb"

    def fallback(self) -> bool:
        return self._fallback

    def _replace_query_for_table(self, table_name: str):
        """
        Build a `SELECT * REPLACE(...) FROM table_name` query for a table
        that converts unsupported column types to varchar columns
        """
        rel = self.conn.view(table_name)
        replaces = []
        for col, type_name in zip(rel.columns, rel.dtypes):
            quoted_col_name = quote_column(col)
            try:
                duckdb_type_name_to_pyarrow_type(type_name)
                # Skip columns with supported types
            except ValueError:
                # Convert unsupported types to strings (except struct)
                if not str(type_name).startswith("STRUCT"):
                    replaces.append(f"{quoted_col_name}::varchar as {quoted_col_name}")

        if replaces:
            replace_csv = ", ".join(replaces)
            return f"SELECT * REPLACE({replace_csv}) FROM {table_name}"
        else:
            return f"SELECT * FROM {table_name}"

    def _schema_for_table(self, table_name: str):
        rel = self.conn.query(f'select * from "{table_name}" limit 1')
        return duckdb_relation_to_schema(rel)

    def tables(self) -> Dict[str, pa.Schema]:
        result = {}
        table_names = self.conn.query(
            "select table_name from information_schema.tables"
        ).to_df()["table_name"].tolist()

        for table_name in table_names:
            if table_name in self._registered_table_schemas:
                # Registered tables are expected to only change when self.register_* is called,
                # so use the cached version
                result[table_name] = self._registered_table_schemas[table_name]
            elif not table_name.startswith(RAW_PREFIX):
                # Dynamically look up schema for tables that are registered with duckdb but not with
                # the self.register_* methods. Skip raw tables
                result[table_name] = self._schema_for_table(table_name)

        return result

    def fetch_query(self, query: str, schema: pa.Schema) -> pa.Table:
        self.logger.info(f"Query:\n{query}\n")
        if self._verbose:
            print(f"DuckDB Query:\n{query}\n")

        # Save query result to temporary view
        tmp_table = "_duckdb_tmp_tbl" + str(uuid.uuid4()).replace("-", "_")
        self.conn.query(query).to_view(tmp_table)

        try:
            # Run replacement query to cast types to match schema
            replace_query = pyarrow_schema_to_select_replace(schema, tmp_table)
            result = self.conn.query(replace_query).to_arrow_table(8096)
        finally:
            # Unregister temporary view
            self.conn.unregister(tmp_table)

        return result

    def _update_temp_names(self, name: str, temporary: bool):
        if temporary:
            self._temp_tables.add(name)
        elif name in self._temp_tables:
            self._temp_tables.remove(name)

    def register_pandas(self, name: str, df: pd.DataFrame, temporary: bool = False):
        # Add _vf_order column to avoid the more expensive operation of computing it with a
        # ROW_NUMBER function in duckdb
        df = df.copy(deep=False)
        df["_vf_order"] = range(0, len(df))

        # Register raw DataFrame under name with prefix
        raw_name = RAW_PREFIX + name
        self.conn.register(raw_name, df)

        # then convert unsupported columns and register result as name
        replace_query = self._replace_query_for_table(raw_name)
        self.conn.query(replace_query).to_view(name)

        self._update_temp_names(name, temporary)
        self._registered_table_schemas[name] = self._schema_for_table(name)

    def register_arrow(self, name: str, table: pa.Table, temporary: bool = False):
        # Register raw table under name with prefix
        raw_name = RAW_PREFIX + name
        self.conn.register(raw_name, table)

        # then convert unsupported columns and register result as name
        replace_query = self._replace_query_for_table(raw_name)
        self.conn.query(replace_query).to_view(name)

        self._update_temp_names(name, temporary)
        self._registered_table_schemas[name] = table.schema

    def register_json(self, name: str, path: str, temporary: bool = False):
        relation = self.conn.read_json(path)
        relation.to_view(name)
        self._update_temp_names(name, temporary)
        self._registered_table_schemas[name] = self._schema_for_table(name)

    def register_csv(self, name: str, path: str, options: CsvReadOptions, temporary: bool = False):
        relation = self.conn.read_csv(
            path,
            header=options.has_header,
            delimiter=options.delimeter,
            all_varchar=True,
        )
        if options.schema is not None:
            replaces = []
            for col_name in options.schema.names:
                field = options.schema.field_by_name(col_name)
                quoted_col_name = quote_column(col_name)
                dtype = field.type
                if dtype == pa.string():
                    # Everything is loaded as string
                    pass
                elif dtype in (pa.float16(), pa.float32(), pa.float64()):
                    replaces.append(f"{quoted_col_name}::double as {quoted_col_name}")
                elif dtype in (pa.int8(), pa.int16(), pa.int32(), pa.int64()):
                    replaces.append(f"{quoted_col_name}::int as {quoted_col_name}")

            if replaces:
                query = f"SELECT * REPLACE({','.join(replaces)}) from _tbl"
                relation = relation.query("_tbl", query)

        relation.to_view(name)
        self._update_temp_names(name, temporary)
        self._registered_table_schemas[name] = self._schema_for_table(name)

    def register_parquet(self, name: str, path: str, temporary: bool = False):
        # Register raw table under name with prefix
        raw_name = RAW_PREFIX + name
        self.conn.read_parquet(path).to_view(raw_name)

        # then convert unsupported columns and register result as name
        replace_query = self._replace_query_for_table(raw_name)
        self.conn.query(replace_query).to_view(name)

        self._update_temp_names(name, temporary)
        self._registered_table_schemas[name] = self._schema_for_table(name)

    def register_arrow_file(self, name: str, path: str, temporary: bool = False):
        arrow_table = pa.feather.read_table(path)
        self.register_arrow(name, arrow_table, temporary)

    def unregister(self, name: str):
        for view_name in [name, RAW_PREFIX + name]:
            self.conn.unregister(view_name)
            if view_name in self._temp_tables:
                self._temp_tables.remove(view_name)
            self._registered_table_schemas.pop(view_name, None)

    def unregister_temporary_tables(self):
        for name in list(self._temp_tables):
            self.conn.unregister(name)
            self._temp_tables.remove(name)


def quote_column(name: str):
    return '"' + name.replace('"', '""') + '"'



# https://altair-viz.github.io/gallery/bar_chart_with_highlighted_segment.html

import altair as alt
import pandas as pd
from vega_datasets import data

source = data.wheat()
threshold = pd.DataFrame([{"threshold": 90}])

bars = alt.Chart(source).mark_bar().encode(
    x="year:O",
    y="wheat:Q",
)

highlight = alt.Chart(source).mark_bar(color="#e45755").encode(
    x='year:O',
    y='baseline:Q',
    y2='wheat:Q'
).transform_filter(
    alt.datum.wheat > 90
).transform_calculate("baseline", "90")

rule = alt.Chart(threshold).mark_rule().encode(
    y='threshold:Q'
)

(bars + highlight + rule).properties(width=500)



# https://altair-viz.github.io/gallery/line_with_log_scale.html

import altair as alt
from vega_datasets import data

source = data.population()

alt.Chart(source).mark_line().encode(
    x='year:O',
    y=alt.Y(
        'sum(people)',
        scale=alt.Scale(type="log")  # Here the scale is applied
    )
)



# https://altair-viz.github.io/gallery/ranged_dot_plot.html

import altair as alt
from vega_datasets import data

source = data.countries.url

chart = alt.layer(
    data=source
).transform_filter(
    filter={"field": 'country',
            "oneOf": ["China", "India", "United States", "Indonesia", "Brazil"]}
).transform_filter(
    filter={'field': 'year',
            "oneOf": [1955, 2000]}
)

chart += alt.Chart().mark_line(color='#db646f').encode(
    x='life_expect:Q',
    y='country:N',
    detail='country:N'
)
# Add points for life expectancy in 1955 & 2000
chart += alt.Chart().mark_point(
    size=100,
    opacity=1,
    filled=True
).encode(
    x='life_expect:Q',
    y='country:N',
    color=alt.Color('year:O',
                    scale=alt.Scale(
                        domain=[1955, 2000],
                        range=['#e6959c', '#911a24']
                    )
                    )
).interactive()

chart



# https://altair-viz.github.io/gallery/scatter_with_rolling_mean.html

import altair as alt
from vega_datasets import data

source = data.seattle_weather()

line = alt.Chart(source).mark_line(
    color='red',
    size=3
).transform_window(
    rolling_mean='mean(temp_max)',
    # sort=[{"field": "date"}],  # Shouldn't need this
    frame=[-15, 15]
).encode(
    x='date:T',
    y='rolling_mean:Q'
)

points = alt.Chart(source).mark_point().encode(
    x='date:T',
    y=alt.Y('temp_max:Q',
            axis=alt.Axis(title='Max Temp'))
)

points + line



# https://altair-viz.github.io/gallery/choropleth.html

import altair as alt
from vega_datasets import data

counties = alt.topo_feature(data.us_10m.url, 'counties')
source = data.unemployment.url

alt.Chart(counties).mark_geoshape().encode(
    color='rate:Q'
).transform_lookup(
    lookup='id',
    from_=alt.LookupData(source, 'id', ['rate'])
).project(
    type='albersUsa'
).properties(
    width=500,
    height=300
)



# https://altair-viz.github.io/gallery/pie_chart.html

import pandas as pd
import altair as alt

source = pd.DataFrame(
    {"category": ["a", "b", "c", "d", "e", "f"], "value": [4, 6, 10, 3, 7, 8]}
)

base = alt.Chart(source).encode(
    theta=alt.Theta("value:Q", stack=True), color=alt.Color("category:N", legend=None)
)

pie = base.mark_arc(outerRadius=120)
text = base.mark_text(radius=140, size=20).encode(text="category:N")

pie + text



# https://altair-viz.github.io/gallery/line_chart_with_datum.html

import altair as alt
from vega_datasets import data

source = data.stocks()

lines = (
    alt.Chart(source)
        .mark_line()
        .encode(x="date", y="price", color="symbol")
)

xrule = (
    alt.Chart()
        .mark_rule(color="cyan", strokeWidth=2)
        .encode(x=alt.datum(alt.DateTime(year=2006, month="November")))
)

yrule = (
    alt.Chart().mark_rule(strokeDash=[12, 6], size=2).encode(y=alt.datum(350))
)


lines + yrule + xrule


# https://altair-viz.github.io/gallery/bar_chart_with_mean_line.html

import altair as alt
from vega_datasets import data

source = data.wheat()

bar = alt.Chart(source).mark_bar().encode(
    x='year:O',
    y='wheat:Q'
)

rule = alt.Chart(source).mark_rule(color='red').encode(
    y='mean(wheat):Q'
)

(bar + rule).properties(width=600)



# https://altair-viz.github.io/gallery/streamgraph.html

import altair as alt
from vega_datasets import data

source = data.unemployment_across_industries.url

alt.Chart(source).mark_area().encode(
    alt.X('yearmonth(date):T',
          axis=alt.Axis(format='%Y', domain=False, tickSize=0)
          ),
    alt.Y('sum(count):Q', stack='center', axis=None),
    alt.Color('series:N',
              scale=alt.Scale(scheme='category20b')
              )
).interactive()



# https://altair-viz.github.io/gallery/line_chart_with_points.html

import altair as alt
import numpy as np
import pandas as pd

x = np.arange(100)
source = pd.DataFrame({
    'x': x,
    'f(x)': np.sin(x / 5)
})

alt.Chart(source).mark_line(
    point=alt.OverlayMarkDef(color="red")
).encode(
    x='x',
    y='f(x)'
)



# https://altair-viz.github.io/gallery/errorbars_with_std.html

import altair as alt
from vega_datasets import data

source = data.barley()

error_bars = alt.Chart(source).mark_errorbar(extent='stdev').encode(
    x=alt.X('yield:Q', scale=alt.Scale(zero=False)),
    y=alt.Y('variety:N')
)

points = alt.Chart(source).mark_point(filled=True, color='black').encode(
    x=alt.X('yield:Q', aggregate='mean'),
    y=alt.Y('variety:N'),
)

error_bars + points



# https://altair-viz.github.io/gallery/ridgeline_plot.html

import altair as alt
from vega_datasets import data

source = "https://raw.githubusercontent.com/vega/vega-datasets/v1.31.1/data/seattle-weather.csv"

step = 20
overlap = 1

alt.Chart(source, height=step).transform_timeunit(
    Month='month(date)'
).transform_joinaggregate(
    mean_temp='mean(temp_max)', groupby=['Month']
).transform_bin(
    ['bin_max', 'bin_min'], 'temp_max'
).transform_aggregate(
    value='count()', groupby=['Month', 'mean_temp', 'bin_min', 'bin_max']
).transform_impute(
    impute='value', groupby=['Month', 'mean_temp'], key='bin_min', value=0
).mark_area(
    interpolate='monotone',
    fillOpacity=0.8,
    stroke='lightgray',
    strokeWidth=0.5
).encode(
    alt.X('bin_min:Q', bin='binned', title='Maximum Daily Temperature (C)'),
    alt.Y(
        'value:Q',
        scale=alt.Scale(range=[step, -step * overlap]),
        axis=None
    ),
    alt.Fill(
        'mean_temp:Q',
        legend=None,
        scale=alt.Scale(domain=[30, 5], scheme='redyellowblue')
    )
).facet(
    row=alt.Row(
        'Month:T',
        title=None,
        header=alt.Header(labelAngle=0, labelAlign='right', format='%B')
    )
).properties(
    title='Seattle Weather',
    bounds='flush'
).configure_facet(
    spacing=0
).configure_view(
    stroke=None
).configure_title(
    anchor='end'
)



# https://altair-viz.github.io/gallery/multi_series_line.html

import altair as alt
from vega_datasets import data

source = data.stocks()

alt.Chart(source).mark_line().encode(
    x='date',
    y='price',
    color='symbol',
    strokeDash='symbol',
)



from pathlib import Path

import pandas as pd
import pyarrow as pa
import pytest
from altair.utils.execeval import eval_block
import vegafusion as vf
from vega_datasets import data
import polars as pl
import altair as alt

pa_major_minor = tuple((int(v) for v in pa.__version__.split(".")[:2]))

here = Path(__file__).parent
altair_mocks_dir = here / "altair_mocks"


def get_connections():
    connections = ["datafusion"]
    try:
        import duckdb
        connections.append("duckdb")
    except ImportError:
        pass

    return connections


@pytest.mark.parametrize(
    "mock_name,expected_len,expected_cols", [
        ("area/cumulative_count", 3201, ["Running_Time_min", "cumulative_count"]),
        ("area/gradient", 68, ["symbol", "date", "price"]),
        ("area/layered", 51, ["year", "source", "net_generation"]),
        ("area/normalized_stacked", 51, ["year", "source", "net_generation_start", "net_generation_end"]),
        ("area/streamgraph", 1708, ["series", "yearmonth_date", "sum_count_start", "sum_count_end"]),
        ("area/trellis", 51, ["year", "source", "net_generation_start", "net_generation_end"]),
        ("area/trellis_sort_array", 492, ["symbol", "date", "row_symbol_sort_index", "price_start", "price_end"]),
        ("bar/diverging_stacked", 40, ["question", "percentage", "percentage_start", "percentage_end"]),
        ("bar/grouped", 12, ["year", "site", "sum_yield"]),
        ("bar/horizontal", 52, ["year", "wheat", "wages"]),
        ("bar/horizontal_grouped", 12, ["year", "site", "sum_yield"]),
        ("bar/horizontal_stacked", 60, ["site", "variety", "sum_yield_start", "sum_yield_end"]),
        ("bar/layered", 51, ["year", "source", "net_generation"]),
        ("bar/normalized_stacked", 60, ["site", "variety", "sum_yield_start", "sum_yield_end"]),
        ("bar/percentage_of_total", 5, ["Activity", "Time", "TotalTime", "PercentOfTotal"]),
        ("bar/sorted", 6, ["site", "sum_yield"]),
        ("bar/stacked", 60, ["site", "variety", "sum_yield_start", "sum_yield_end"]),
        ("bar/stacked_with_sorted_segments", 60, ["site", "variety", "sum_yield_start", "sum_yield_end"]),
        ("bar/trellis_compact", 27, ["a", "b", "c", "p"]),
        ("bar/trellis_stacked", 120, ["yield", "variety", "year", "yield_start", "yield_end"]),
        ("bar/with_highlighted_bar", 52, ["year", "wheat", "wages"]),
        ("bar/with_negative_values", 120, ["month", "nonfarm_change"]),
        ("bar/with_rounded_edges", 53, ["weather", "month_date", "__count_start", "__count_end"]),
        ("casestudy/anscombe_plot", 44, ["Series", "X", "Y"]),
        ("casestudy/beckers_barley_trellis_plot", 120, ["yield", "variety", "year", "site"]),
        ("casestudy/gapminder_bubble_plot", 187, ["country", "health", "population"]),
        ("casestudy/iowa_electricity", 51, ["year", "net_generation_start", "net_generation_end"]),
        ("casestudy/isotype", 37, ["country", "animal", "x"]),
        ("casestudy/natural_disasters", 686, ["Entity", "Year", "Deaths"]),
        ("casestudy/top_k_items", 9, ["Title", "IMDB_Rating_start", "IMDB_Rating_end"]),
        ("casestudy/top_k_letters", 9, ["letters", "count", "rank"]),
        ("casestudy/top_k_with_others", 10, ["ranked_director", "mean_aggregate_gross"]),
        ("casestudy/us_population_over_time_facet", 285, ["age", "year", "sum_people"]),
        ("casestudy/window_rank", 12, ["team", "matchday", "rank"]),
        ("circular/donut", 6, ["category", "value_start", "value_end"]),
        ("circular/pie", 6, ["category", "value_start", "value_end"]),
        ("histogram/trellis", 20, ["Origin", "__count", "bin_maxbins_10_Horsepower_end"]),
        ("histogram/layered", 113, ["Experiment", "__count", "bin_maxbins_100_Measurement"]),
        ("interactive/brush", 392, ["Name", "Cylinders", "Origin"]),
        ("interactive/casestudy-us_population_over_time", 38, ["year", "age", "sex", "people"]),
        ("interactive/casestudy-weather_heatmap", 365, ["monthdate_date", "date_date", "max_temp"]),
        ("interactive/legend", 1708, ["yearmonth_date", "sum_count_start", "sum_count_end"]),
        ("interactive/other-image_tooltip", 2, ["a", "b", "image"]),
        ("interactive/scatter-href", 392, ["Name", "Horsepower", "url"]),
        ("interactive/scatter_plot", 392, ["Name", "Horsepower", "Year"]),
        ("line/bump_chart", 100, ["symbol", "rank", "yearmonth_date_end"]),
        ("line/filled_step_chart", 68, ["symbol", "date", "price"]),
        ("line/multi_series", 560, ["symbol", "date", "price"]),
        ("line/percent_axis", 30, ["job", "sex", "year", "count", "perc"]),
        ("line/slope_graph", 12, ["site", "year", "median_yield"]),
        ("line/step_chart", 68, ["symbol", "date", "price"]),
        ("line/with_cumsum", 52, ["year", "wheat", "cumulative_wheat"]),
        ("line/with_generator", 256, ["x", "sin", "cos", "key", "value"]),
        ("line/with_logarithmic_scale", 15, ["year", "sum_people"]),
        ("line/with_points", 100, ["x", "f(x)"]),
        ("other/beckers_barley_wrapped_facet", 120, ["variety", "site", "median_yield"]),
        ("other/binned_heatmap", 378, ["__count", "bin_maxbins_60_IMDB_Rating_end"]),
        ("other/boxplot", 19, ["age", "mid_box_people"]),
        ("other/comet_chart", 120, ["variety", "1932", "delta"]),
        ("other/gantt_chart", 3, ["task", "start", "end"]),
        ("other/hexbins", 84, ["month_date", "xFeaturePos", "mean_temp_max"]),
        ("other/isotype_grid", 100, ["id", "row", "col"]),
        ("other/multiple_marks", 560, ["symbol", "date", "price"]),
        ("other/stem_and_leaf", 100, ["samples", "stem", "leaf", "position"]),
        ("other/wilkinson_dot_plot", 21, ["data", "id"]),
        ("other/parallel_coordinates", 600, ["sepalWidth", "index", "key", "value"]),
        ("other/normed_parallel_coordinates", 600, ["sepalWidth", "minmax_value", "mid"]),
        ("other/ridgeline_plot", 108, ["Month", "mean_temp", "value"]),
        ("scatter/binned", 64, ["__count", "bin_maxbins_10_Rotten_Tomatoes_Rating_end"]),
        ("scatter/bubble_plot", 392, ["Name", "Cylinders", "Origin"]),
        ("scatter/connected", 55, ["side", "year", "miles", "gas"]),
        ("scatter/multifeature", 150, ["sepalLength", "petalLength", "species"]),
        ("scatter/table_bubble_plot_github", 168, ["hours_time", "day_time", "sum_count"]),
        ("scatter/trellis", 392, ["Name", "Cylinders", "Year"]),
        ("simple/bar_chart", 9, ["a", "b"]),
        ("simple/heatmap", 100, ["x", "y", "z"]),
        ("simple/line_chart", 100, ["x", "f(x)"]),
        ("simple/scatter_tooltips", 392, ["Name", "Cylinders", "Year"]),
        ("simple/stacked_bar_chart", 51, ["year", "source", "net_generation_end"]),
        ("simple/strip_chart", 400, ["Name", "Cylinders", "Origin"]),
    ]
)
@pytest.mark.parametrize("connection", get_connections())
def test_transformed_data_for_mock(mock_name, expected_len, expected_cols, connection):
    vf.runtime.set_connection(connection)

    mock_path = altair_mocks_dir / mock_name / "mock.py"
    mock_src = mock_path.read_text("utf8")
    chart = eval_block(mock_src)
    df = vf.transformed_data(chart)

    # Check that a DataFrame was returned
    assert isinstance(df, pd.DataFrame)

    # Check that the expected columns are present
    assert set(expected_cols).issubset(set(df.columns))

    # Check that datetime columns have local timezone
    for dtype in df.dtypes.values:
        if dtype.kind == "M":
            assert str(dtype.tz) == vf.get_local_tz()

    # Check expected length
    assert len(df) == expected_len


@pytest.mark.parametrize(
    "mock_name,expected_lens,all_expected_cols", [
        ("area/horizon_graph", [20, 20], [["x", "y"], ["x", "y", "ny"]]),
        ("bar/and_tick_chart", [7, 7], [["goal", "score_start"], ["project", "goal"]]),
        ("bar/stacked_with_text_overlay", [60, 60], [["site", "sum_yield_start"], ["variety", "sum_yield_end"]]),
        ("bar/with_labels", [52, 52], [["wages", "wheat_start"], ["wheat", "wages"]]),
        ("bar/with_line_at_mean", [52, 1], [["wages", "wheat_start"], ["mean_wheat"]]),
        ("bar/with_line_on_dual_axis", [52, 52], [["wages", "wheat_start"], ["wheat", "wages"]]),
        ("bar/with_rolling_mean", [52, 52], [["wages", "wheat_start"], ["wheat", "wages"]]),
        ("casestudy/co2_concentration", [713, 7, 7], [["year", "decade"], ["scaled_date", "first_date"], ["end"]]),
        ("casestudy/falkensee", [2, 38, 38], [["event", "start"], ["population", "year"], ["year"]]),
        ("casestudy/us_employment", [120, 1, 2], [["construction"], ["president", "end"], ["start"]]),
        ("casestudy/wheat_wages", [52, 52, 52, 52], [["wheat"], ["year_end"], ["year"], ["year"]]),
        ("histogram/with_a_global_mean_overlay", [9, 1], [["bin_maxbins_10_IMDB_Rating_end"], ["mean_IMDB_Rating"]]),
        ("interactive/area-interval_selection", [123, 123], [["price_start"], ["price_end"]]),
        ("interactive/casestudy-seattle_weather_interactive", [1461, 5], [["monthdate_date"], ["__count"]]),
        ("interactive/casestudy-us_population_pyramid_over_time", [19, 38, 19], [["sum_people"], ["people"], ["sum_people_end"]]),
        ("interactive/cross_highlight", [64, 64, 13], [["__count"], ["__count"], ["__count"]]),
        ("interactive/histogram-responsive", [20, 20], [["__count"], ["__count"]]),
        ("interactive/multiline_highlight", [560, 560], [["price"], ["price"]]),
        ("interactive/multiline_tooltip", [300, 300, 300], [["x"], ["y"], ["category"]]),
        ("interactive/scatter-with_linked_table", [392, 19, 19, 19], [["Year"], ["rank"], ["rank"], ["rank"]]),
        ("interactive/scatter-with_minimap", [1461, 1461], [["weather"], ["weather"]]),
        ("interactive/scatter_with_layered_histogram", [2, 19], [["mean_height"], ["bin_step_5_age"]]),
        ("interactive/select_detail", [20, 1000], [["mean_y"], ["value"]]),
        ("interactive/select_mark_area", [122, 122], [["sum_count"], ["yearmonth_date"]]),
        ("interactive/selection_histogram", [392, 3], [["Cylinders"], ["__count"]]),
        ("interactive/selection_layer_bar_month", [12, 1], [["mean_precipitation"], ["mean_precipitation"]]),
        ("line/layer_line_color_rule", [560, 5], [["symbol"], ["average_price"]]),
        ("other/bar_chart_with_highlighted_segment", [52, 1, 1], [["wheat_start"], ["baseline"], ["threshold"]]),
        ("other/candlestick_chart", [44, 44], [["ret"], ["signal"]]),
        ("other/errorbars_with_std", [10, 10], [["mean_yield"], ["variety"]]),
        ("other/layered_chart_with_dual_axis", [12, 12], [["average_temp_max"], ["average_temp_max"]]),
        ("other/layered_heatmap_text", [9, 9], [["Origin"], ["num_cars"]]),
        ("other/ranged_dot_plot", [10, 10], [["life_expect"], ["country"]]),
        ("other/scatter_marginal_hist", [34, 150, 27], [["__count"], ["species"], ["__count"]]),
        ("scatter/dot_dash_plot", [400, 392, 398], [["Cylinders"], ["Cylinders"], ["Cylinders"]]),
        ("scatter/with_errorbars", [5, 5], [["ymin"], ["upper_ymin"]]),
        ("scatter/with_labels", [5, 5], [["x"], ["label"]]),
        ("scatter/with_rolling_mean", [1461, 1461], [["precipitation"], ["rolling_mean"]]),
    ]
)
@pytest.mark.parametrize("connection", get_connections())
def test_multi_transformed_data_for_mock(mock_name, expected_lens, all_expected_cols, connection):
    vf.runtime.set_connection(connection)
    mock_path = altair_mocks_dir / mock_name / "mock.py"
    mock_src = mock_path.read_text("utf8")
    chart = eval_block(mock_src)
    dfs = vf.transformed_data(chart)

    for df, expected_len, expected_cols in zip(dfs, expected_lens, all_expected_cols):
        # Check that a DataFrame was returned
        assert isinstance(df, pd.DataFrame)

        # Check that the expected columns are present
        assert set(expected_cols).issubset(set(df.columns))

        # Check that datetime columns have local timezone
        for dtype in df.dtypes.values:
            if dtype.kind == "M":
                assert str(dtype.tz) == vf.get_local_tz()

        # Check expected length
        assert len(df) == expected_len


def test_transformed_data_exclude():
    source = data.wheat()
    bar = alt.Chart(source).mark_bar().encode(x="year:O", y="wheat:Q")
    rule = alt.Chart(source).mark_rule(color="red").encode(y="mean(wheat):Q")
    some_annotation = (
        alt.Chart(name="some_annotation")
        .mark_text(fontWeight="bold")
        .encode(text=alt.value("Just some text"), y=alt.datum(85), x=alt.value(200))
    )

    chart = (bar + rule + some_annotation).properties(width=600)
    datasets = vf.transformed_data(chart, exclude=["some_annotation"])

    assert len(datasets) == 2
    assert len(datasets[0]) == 52
    assert "wheat_start" in datasets[0]
    assert len(datasets[1]) == 1
    assert "mean_wheat" in datasets[1]


@pytest.mark.skipif(pa_major_minor < (11, 0), reason="pyarrow 11+ required")
def test_gh_286():
    # https://github.com/hex-inc/vegafusion/issues/286
    vf.runtime.set_connection("datafusion")
    source = pl.from_pandas(data.seattle_weather())

    chart = alt.Chart(source).mark_bar(
        cornerRadiusTopLeft=3,
        cornerRadiusTopRight=3
    ).encode(
        x='month(date):O',
        y='count():Q',
        color='weather:N'
    )
    transformed = vf.transformed_data(chart)
    assert isinstance(transformed, pl.DataFrame)
    assert len(transformed) == 53


@pytest.mark.parametrize("connection", get_connections())
def test_categorical_columns(connection):
    vf.runtime.set_connection(connection)

    df = pd.DataFrame({
        "a": [0, 1, 2, 3, 4, 5],
        "categorical": pd.Categorical.from_codes([0, 1, 0, 1, 1, 0], ["A", "BB"])
    })

    chart = alt.Chart(df).mark_bar().encode(
        alt.X("categorical:N"), alt.Y("sum(a):Q")
    )
    transformed = vf.transformed_data(chart)
    expected = pd.DataFrame({"categorical": ["A", "BB"], "sum_a": [7, 8]})
    pd.testing.assert_frame_equal(transformed, expected)


# https://altair-viz.github.io/gallery/filled_step_chart.html

import altair as alt
from vega_datasets import data

source = data.stocks()

alt.Chart(source).mark_area(
    color="lightblue",
    interpolate='step-after',
    line=True
).encode(
    x='date',
    y='price'
).transform_filter(alt.datum.symbol == 'GOOG')



# https://altair-viz.github.io/gallery/scatter_linked_brush.html

import altair as alt
from vega_datasets import data

source = data.cars()

brush = alt.selection(type='interval', resolve='global')

base = alt.Chart(source).mark_point().encode(
    y='Miles_per_Gallon',
    color=alt.condition(brush, 'Origin', alt.ColorValue('gray')),
).add_selection(
    brush
).properties(
    width=200,
    height=250
)

base.encode(x='Horsepower') | base.encode(x='Acceleration')



# https://altair-viz.github.io/gallery/line_with_ci.html
# extent changed for "ci" to "stdev" to be deterministic

import altair as alt
from vega_datasets import data

source = data.cars()

line = alt.Chart(source).mark_line().encode(
    x='Year',
    y='mean(Miles_per_Gallon)'
)

band = alt.Chart(source).mark_errorband(extent='stdev').encode(
    x='Year',
    y=alt.Y('Miles_per_Gallon', title='Miles/Gallon'),
)

band + line



import altair as alt
from vega_datasets import data

source = data.stocks()

highlight = alt.selection_point(on='mouseover',
                                fields=['symbol'], nearest=True)

base = alt.Chart(source).encode(
    x='date:T',
    y='price:Q',
    color='symbol:N'
)

points = base.mark_circle().encode(
    opacity=alt.value(0)
).add_params(
    highlight
).properties(
    width=600
)

lines = base.mark_line().encode(
    size=alt.condition(~highlight, alt.value(1), alt.value(3))
)

points + lines


from abc import ABC, abstractmethod
import pyarrow as pa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dfi import SqlDatasetDataFrame


class SqlDataset(ABC):
    """
    Python interface for VegaFusion Sql Dataset
    """

    def dialect(self) -> str:
        """
        Returns SQL dialect accepted by the connection

        :return: One of:
         - "athena"
         - "bigquery"
         - "clickhouse"
         - "databricks"
         - "datafusion"
         - "duckdb"
         - "mysql"
         - "postgres"
         - "redshift"
         - "snowflake"
        """
        raise NotImplementedError()

    @abstractmethod
    def table_name(self) -> str:
        """Name of source table for use in queries"""
        raise NotImplementedError()

    @abstractmethod
    def table_schema(self) -> pa.Schema:
        """Schema of source table"""
        raise NotImplementedError()

    @abstractmethod
    def fetch_query(self, query: str, schema: pa.Schema) -> pa.Table:
        """
        Returns the result of evaluating the requested query. The resulting pa.Table
        should have a schema matching the provided schema

        :param query: SQL query string
        :param schema: expected pyarrow Schema of resulting pyarrow Table
        :return: pyarrow Table with query results
        """
        raise NotImplementedError()

    def fallback(self) -> bool:
        """
        Whether VegaFusion should fall back to the built-in DataFusion connection
        when SQL is encountered that is not supported by this connection's SQL dialect

        :return: bool
        """
        return True

    def main_thread(self) -> bool:
        """
        True if the dataset must be evaluated run on the main thread.
        This blocks multithreaded parallelization, but is sometimes required

        :return: bool
        """
        return True

    def __dataframe__(
        self, nan_as_null: bool = False, allow_copy: bool = True, **kwargs
    ) -> "SqlDatasetDataFrame":
        """DataFrame interchange protocol support"""
        from .dfi import SqlDatasetDataFrame
        return SqlDatasetDataFrame(self, nan_as_null=nan_as_null, allow_copy=allow_copy)



# https://altair-viz.github.io/gallery/choropleth_repeat.html

import altair as alt
from vega_datasets import data

states = alt.topo_feature(data.us_10m.url, 'states')
source = data.population_engineers_hurricanes.url
variable_list = ['population', 'engineers', 'hurricanes']

alt.Chart(states).mark_geoshape().encode(
    alt.Color(alt.repeat('row'), type='quantitative')
).transform_lookup(
    lookup='id',
    from_=alt.LookupData(source, 'id', variable_list)
).properties(
    width=500,
    height=150
).project(
    type='albersUsa'
).repeat(
    row=variable_list
).resolve_scale(
    color='independent'
)



# https://altair-viz.github.io/gallery/interactive_brush.html

import altair as alt
from vega_datasets import data

source = data.cars()
brush = alt.selection(type='interval')

alt.Chart(source).mark_point().encode(
    x='Horsepower:Q',
    y='Miles_per_Gallon:Q',
    color=alt.condition(brush, 'Cylinders:O', alt.value('grey')),
).add_selection(brush)



# https://altair-viz.github.io/gallery/grouped_bar_chart_with_error_bars.html

import altair as alt
from vega_datasets import data

source = data.barley()

bars = alt.Chart().mark_bar().encode(
    x='year:O',
    y=alt.Y('mean(yield):Q', title='Mean Yield'),
    color='year:N',
)

error_bars = alt.Chart().mark_errorbar(extent='ci').encode(
    x='year:O',
    y='yield:Q'
)

alt.layer(bars, error_bars, data=source).facet(
    column='site:N'
)


# https://altair-viz.github.io/gallery/natural_disasters.html

import altair as alt
from vega_datasets import data

source = "https://raw.githubusercontent.com/vega/vega-datasets/main/data/disasters.csv"

alt.Chart(source).mark_circle(
    opacity=0.8,
    stroke='black',
    strokeWidth=1
).encode(
    alt.X('Year:O', axis=alt.Axis(labelAngle=0)),
    alt.Y('Entity:N'),
    alt.Size('Deaths:Q',
             scale=alt.Scale(range=[0, 4000]),
             legend=alt.Legend(title='Annual Global Deaths')
             ),
    alt.Color('Entity:N', legend=None)
).properties(
    width=300,
    height=300
).transform_filter(
    alt.datum.Entity != 'All natural disasters'
)




# https://altair-viz.github.io/gallery/donut_chart.html

import pandas as pd
import altair as alt

source = pd.DataFrame({"category": [1, 2, 3, 4, 5, 6], "value": [4, 6, 10, 3, 7, 8]})

alt.Chart(source).mark_arc(innerRadius=50).encode(
    theta=alt.Theta(field="value", type="quantitative"),
    color=alt.Color(field="category", type="nominal"),
)



# https://altair-viz.github.io/gallery/multifeature_scatter_plot.html

import altair as alt
from vega_datasets import data

source = data.iris()

alt.Chart(source).mark_circle().encode(
    alt.X('sepalLength', scale=alt.Scale(zero=False)),
    alt.Y('sepalWidth', scale=alt.Scale(zero=False, padding=1)),
    color='species',
    size='petalWidth'
)



# https://altair-viz.github.io/gallery/bump_chart.html
# Width set to 500

import altair as alt
from vega_datasets import data
import pandas as pd

stocks = data.stocks()
source = stocks.groupby([pd.Grouper(key="date", freq="6M"),"symbol"]).mean().reset_index()

alt.Chart(source).mark_line(point = True).encode(
    x = alt.X("date:O", timeUnit="yearmonth", title="date"),
    y="rank:O",
    color=alt.Color("symbol:N")
).transform_window(
    rank="rank()",
    sort=[alt.SortField("price", order="descending")],
    groupby=["date"]
).properties(
    title="Bump Chart for Stock Prices",
    width=500,
    height=150,
)


from vegafusion.dataset.duckdb import DuckDbDataset
import duckdb
import pytest


def test_sql_dataset_dfi():
    try:
        import pyarrow.interchange as pi
    except ImportError:
        pytest.skip("DataFrame interface protocol requires pyarrow 11.0.0 or later")

    rel = duckdb.query("SELECT 1 as a")
    dataset = DuckDbDataset(rel)
    table = pi.from_dataframe(dataset)
    assert table.num_rows == 1
    assert table.column_names == ["a"]



# https://altair-viz.github.io/gallery/bubble_plot.html

import altair as alt
from vega_datasets import data

source = data.cars()

alt.Chart(source).mark_point().encode(
    x='Horsepower',
    y='Miles_per_Gallon',
    size='Acceleration'
)



# https://altair-viz.github.io/gallery/weather_heatmap.html

import altair as alt
from vega_datasets import data

# Since the data is more than 5,000 rows we'll import it from a URL
source = "https://raw.githubusercontent.com/vega/vega-datasets/v1.31.1/data/seattle-temps.csv"

alt.Chart(
    source,
    title="2010 Daily High Temperature (F) in Seattle, WA"
).mark_rect().encode(
    x='date(date):O',
    y='month(date):O',
    color=alt.Color('max(temp):Q', scale=alt.Scale(scheme="inferno")),
    tooltip=[
        alt.Tooltip('monthdate(date):T', title='Date'),
        alt.Tooltip('max(temp):Q', title='Max Temp')
    ]
).properties(width=550)



# https://altair-viz.github.io/gallery/scatter_tooltips.html

import altair as alt
from vega_datasets import data

source = data.cars()

alt.Chart(source).mark_circle(size=60).encode(
    x='Horsepower',
    y='Miles_per_Gallon',
    color='Origin',
    tooltip=['Name', 'Origin', 'Horsepower', 'Miles_per_Gallon']
).interactive()



# https://altair-viz.github.io/gallery/bar_chart_with_negatives.html

import altair as alt
from vega_datasets import data

source = data.us_employment()

alt.Chart(source).mark_bar().encode(
    x="month:T",
    y="nonfarm_change:Q",
    color=alt.condition(
        alt.datum.nonfarm_change > 0,
        alt.value("steelblue"),  # The positive color
        alt.value("orange")  # The negative color
    )
).properties(width=600)


# https://altair-viz.github.io/gallery/step_chart.html

import altair as alt
from vega_datasets import data

source = data.stocks()

alt.Chart(source).mark_line(interpolate='step-after').encode(
    x='date',
    y='price'
).transform_filter(
    alt.datum.symbol == 'GOOG'
)



# https://altair-viz.github.io/gallery/sorted_error_bars_with_ci.html
# Padding updates to make size deterministic even when data is not

import altair as alt
from vega_datasets import data

source = data.barley()

points = alt.Chart(source).mark_point(
    filled=True,
    color='black'
).encode(
    x=alt.X('mean(yield)', title='Barley Yield'),
    y=alt.Y(
        'variety',
        sort=alt.EncodingSortField(
            field='yield',
            op='mean',
            order='descending'
        )
    )
).properties(
    width=400,
    height=250
)

error_bars = points.mark_rule().encode(
    x='ci0(yield)',
    x2='ci1(yield)',
)

(points + error_bars).properties(
    padding={"left": 50, "top": 5, "right": 50, "bottom": 5}
)



# https://altair-viz.github.io/gallery/normed_parallel_coordinates.html

import altair as alt
from vega_datasets import data
from altair import datum

source = data.iris()

alt.Chart(source).transform_window(
    index='count()'
).transform_fold(
    ['petalLength', 'petalWidth', 'sepalLength', 'sepalWidth']
).transform_joinaggregate(
    min='min(value)',
    max='max(value)',
    groupby=['key']
).transform_calculate(
    minmax_value=(datum.value-datum.min)/(datum.max-datum.min),
    mid=(datum.min+datum.max)/2
).mark_line().encode(
    x='key:N',
    y='minmax_value:Q',
    color='species:N',
    detail='index:N',
    opacity=alt.value(0.5)
).properties(width=500)



# https://altair-viz.github.io/gallery/trellis_area.html

import altair as alt
from vega_datasets import data

source = data.iowa_electricity()

alt.Chart(source).mark_area().encode(
    x="year:T",
    y="net_generation:Q",
    color="source:N",
    row="source:N"
).properties(
    height=100
)



import altair as alt
import pandas as pd
import numpy as np

# generate fake data
source = pd.DataFrame({
    'gender': ['M']*1000 + ['F']*1000,
    'height':np.concatenate((
        np.random.normal(69, 7, 1000), np.random.normal(64, 6, 1000)
    )),
    'weight': np.concatenate((
        np.random.normal(195.8, 144, 1000), np.random.normal(167, 100, 1000)
    )),
    'age': np.concatenate((
        np.random.normal(45, 8, 1000), np.random.normal(51, 6, 1000)
    ))
})

selector = alt.selection_point(fields=['gender'])

color_scale = alt.Scale(domain=['M', 'F'],
                        range=['#1FC3AA', '#8624F5'])

base = alt.Chart(source).properties(
    width=250,
    height=250
).add_params(selector)

points = base.mark_point(filled=True, size=200).encode(
    x=alt.X('mean(height):Q').scale(domain=[0,84]),
    y=alt.Y('mean(weight):Q').scale(domain=[0,250]),
    color=alt.condition(
        selector,
        'gender:N',
        alt.value('lightgray'),
        scale=color_scale),
)

hists = base.mark_bar(opacity=0.5, thickness=100).encode(
    x=alt.X('age')
    .bin(step=5) # step keeps bin size the same
    .scale(domain=[0,100]),
    y=alt.Y('count()')
    .stack(None)
    .scale(domain=[0,350]),
    color=alt.Color('gender:N', scale=color_scale)
).transform_filter(
    selector
)

points | hists


# https://altair-viz.github.io/gallery/area_chart_gradient.html

import altair as alt
from vega_datasets import data

source = data.cars()

alt.Chart(source).mark_circle().encode(
    alt.X(alt.repeat("column"), type='quantitative'),
    alt.Y(alt.repeat("row"), type='quantitative'),
    color='Origin:N'
).properties(
    width=120,
    height=120
).repeat(
    row=['Horsepower', 'Acceleration', 'Miles_per_Gallon'],
    column=['Miles_per_Gallon', 'Acceleration', 'Horsepower']
).interactive()


# https://altair-viz.github.io/gallery/scatter_with_loess.html

import altair as alt
import pandas as pd
import numpy as np

np.random.seed(1)

source = pd.DataFrame({
    'x': np.arange(100),
    'A': np.random.randn(100).cumsum(),
    'B': np.random.randn(100).cumsum(),
    'C': np.random.randn(100).cumsum(),
})

base = alt.Chart(source).mark_circle(opacity=0.5).transform_fold(
    fold=['A', 'B', 'C'],
    as_=['category', 'y']
).encode(
    alt.X('x:Q'),
    alt.Y('y:Q'),
    alt.Color('category:N')
)

base + base.transform_loess('x', 'y', groupby=['category']).mark_line(size=4)



import vegafusion as vf
import pandas as pd
import json


def test_input_utc():
    (pre_transformed, warnings) = vf.runtime.pre_transform_spec(
        input_spec(), "UTC", "UTC", inline_datasets={"seattle_weather": load_dataset()}
    )
    print(json.dumps(pre_transformed, indent=2))
    expected = expected_spec()
    assert pre_transformed == expected


def load_dataset():
    """
    Load seattle-weather dataset with the date column localized as UTC
    """
    seattle_weather = pd.read_csv("https://raw.githubusercontent.com/vega/vega-datasets/next/data/seattle-weather.csv")
    seattle_weather["date"] = pd.to_datetime(seattle_weather["date"])
    seattle_weather = seattle_weather.set_index("date").tz_localize("UTC").reset_index()
    return seattle_weather


def input_spec():
    return r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "height": 200,
  "style": "cell",
  "data": [
    {
      "name": "source_0",
      "url": "vegafusion+dataset://seattle_weather",
      "format": {"type": "csv", "parse": {"date": "date"}},
      "transform": [
        {
          "field": "date",
          "type": "timeunit",
          "units": ["month"],
          "as": ["month_date", "month_date_end"]
        },
        {
          "type": "aggregate",
          "groupby": ["month_date", "weather"],
          "ops": ["count"],
          "fields": [null],
          "as": ["__count"]
        },
        {
          "type": "stack",
          "groupby": ["month_date"],
          "field": "__count",
          "sort": {"field": ["weather"], "order": ["descending"]},
          "as": ["__count_start", "__count_end"],
          "offset": "zero"
        },
        {
            "type": "collect",
            "sort": {"field": ["month_date", "weather"]}
        }
      ]
    }
  ],
  "signals": [
    {"name": "x_step", "value": 20},
    {
      "name": "width",
      "update": "bandspace(domain('x').length, 0.1, 0.05) * x_step"
    }
  ],
  "marks": [
    {
      "name": "marks",
      "type": "rect",
      "style": ["bar"],
      "from": {"data": "source_0"},
      "encode": {
        "update": {
          "fill": {"scale": "color", "field": "weather"},
          "ariaRoleDescription": {"value": "bar"},
          "description": {
            "signal": "\"Month of the year: \" + (timeFormat(datum[\"month_date\"], timeUnitSpecifier([\"month\"], {\"year-month\":\"%b %Y \",\"year-month-date\":\"%b %d, %Y \"}))) + \"; Count of Records: \" + (format(datum[\"__count\"], \"\")) + \"; Weather type: \" + (isValid(datum[\"weather\"]) ? datum[\"weather\"] : \"\"+datum[\"weather\"])"
          },
          "x": {"scale": "x", "field": "month_date"},
          "width": {"signal": "max(0.25, bandwidth('x'))"},
          "y": {"scale": "y", "field": "__count_end"},
          "y2": {"scale": "y", "field": "__count_start"}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "band",
      "domain": {"data": "source_0", "field": "month_date", "sort": true},
      "range": {"step": {"signal": "x_step"}},
      "paddingInner": 0.1,
      "paddingOuter": 0.05
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {
        "data": "source_0",
        "fields": ["__count_start", "__count_end"]
      },
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    },
    {
      "name": "color",
      "type": "ordinal",
      "domain": ["sun", "fog", "drizzle", "rain", "snow"],
      "range": ["#e7ba52", "#c7c7c7", "#aec7e8", "#1f77b4", "#9467bd"]
    }
  ],
  "axes": [
    {
      "scale": "y",
      "orient": "left",
      "gridScale": "x",
      "grid": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "x",
      "orient": "bottom",
      "grid": false,
      "title": "Month of the year",
      "format": {
        "signal": "timeUnitSpecifier([\"month\"], {\"year-month\":\"%b %Y \",\"year-month-date\":\"%b %d, %Y \"})"
      },
      "formatType": "time",
      "labelOverlap": true,
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": false,
      "title": "Count of Records",
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "zindex": 0
    }
  ],
  "legends": [
    {"title": "Weather type", "fill": "color", "symbolType": "square"}
  ]
}
"""


def expected_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "data": [
    {
      "name": "source_0",
      "values": [
        {
          "__count": 10,
          "__count_end": 124,
          "__count_start": 114,
          "month_date": "2012-01-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 17,
          "__count_end": 114,
          "__count_start": 97,
          "month_date": "2012-01-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 61,
          "__count_end": 97,
          "__count_start": 36,
          "month_date": "2012-01-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 8,
          "__count_end": 36,
          "__count_start": 28,
          "month_date": "2012-01-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 28,
          "__count_end": 28,
          "__count_start": 0,
          "month_date": "2012-01-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 4,
          "__count_end": 113,
          "__count_start": 109,
          "month_date": "2012-02-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 3,
          "__count_end": 109,
          "__count_start": 106,
          "month_date": "2012-02-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 75,
          "__count_end": 106,
          "__count_start": 31,
          "month_date": "2012-02-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 4,
          "__count_end": 31,
          "__count_start": 27,
          "month_date": "2012-02-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 27,
          "__count_end": 27,
          "__count_start": 0,
          "month_date": "2012-02-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 3,
          "__count_end": 124,
          "__count_start": 121,
          "month_date": "2012-03-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 6,
          "__count_end": 121,
          "__count_start": 115,
          "month_date": "2012-03-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 73,
          "__count_end": 115,
          "__count_start": 42,
          "month_date": "2012-03-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 6,
          "__count_end": 42,
          "__count_start": 36,
          "month_date": "2012-03-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 36,
          "__count_end": 36,
          "__count_start": 0,
          "month_date": "2012-03-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 3,
          "__count_end": 120,
          "__count_start": 117,
          "month_date": "2012-04-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 3,
          "__count_end": 117,
          "__count_start": 114,
          "month_date": "2012-04-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 61,
          "__count_end": 114,
          "__count_start": 53,
          "month_date": "2012-04-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 1,
          "__count_end": 53,
          "__count_start": 52,
          "month_date": "2012-04-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 52,
          "__count_end": 52,
          "__count_start": 0,
          "month_date": "2012-04-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 1,
          "__count_end": 124,
          "__count_start": 123,
          "month_date": "2012-05-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 5,
          "__count_end": 123,
          "__count_start": 118,
          "month_date": "2012-05-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 40,
          "__count_end": 118,
          "__count_start": 78,
          "month_date": "2012-05-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 78,
          "__count_end": 78,
          "__count_start": 0,
          "month_date": "2012-05-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 2,
          "__count_end": 120,
          "__count_start": 118,
          "month_date": "2012-06-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 1,
          "__count_end": 118,
          "__count_start": 117,
          "month_date": "2012-06-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 42,
          "__count_end": 117,
          "__count_start": 75,
          "month_date": "2012-06-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 75,
          "__count_end": 75,
          "__count_start": 0,
          "month_date": "2012-06-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 8,
          "__count_end": 124,
          "__count_start": 116,
          "month_date": "2012-07-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 10,
          "__count_end": 116,
          "__count_start": 106,
          "month_date": "2012-07-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 16,
          "__count_end": 106,
          "__count_start": 90,
          "month_date": "2012-07-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 90,
          "__count_end": 90,
          "__count_start": 0,
          "month_date": "2012-07-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 8,
          "__count_end": 124,
          "__count_start": 116,
          "month_date": "2012-08-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 6,
          "__count_end": 116,
          "__count_start": 110,
          "month_date": "2012-08-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 24,
          "__count_end": 110,
          "__count_start": 86,
          "month_date": "2012-08-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 86,
          "__count_end": 86,
          "__count_start": 0,
          "month_date": "2012-08-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 5,
          "__count_end": 120,
          "__count_start": 115,
          "month_date": "2012-09-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 14,
          "__count_end": 115,
          "__count_start": 101,
          "month_date": "2012-09-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 36,
          "__count_end": 101,
          "__count_start": 65,
          "month_date": "2012-09-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 65,
          "__count_end": 65,
          "__count_start": 0,
          "month_date": "2012-09-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 4,
          "__count_end": 124,
          "__count_start": 120,
          "month_date": "2012-10-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 19,
          "__count_end": 120,
          "__count_start": 101,
          "month_date": "2012-10-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 62,
          "__count_end": 101,
          "__count_start": 39,
          "month_date": "2012-10-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 39,
          "__count_end": 39,
          "__count_start": 0,
          "month_date": "2012-10-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 3,
          "__count_end": 120,
          "__count_start": 117,
          "month_date": "2012-11-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 9,
          "__count_end": 117,
          "__count_start": 108,
          "month_date": "2012-11-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 75,
          "__count_end": 108,
          "__count_start": 33,
          "month_date": "2012-11-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 1,
          "__count_end": 33,
          "__count_start": 32,
          "month_date": "2012-11-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 32,
          "__count_end": 32,
          "__count_start": 0,
          "month_date": "2012-11-01T00:00:00.000",
          "weather": "sun"
        },
        {
          "__count": 2,
          "__count_end": 124,
          "__count_start": 122,
          "month_date": "2012-12-01T00:00:00.000",
          "weather": "drizzle"
        },
        {
          "__count": 8,
          "__count_end": 122,
          "__count_start": 114,
          "month_date": "2012-12-01T00:00:00.000",
          "weather": "fog"
        },
        {
          "__count": 76,
          "__count_end": 114,
          "__count_start": 38,
          "month_date": "2012-12-01T00:00:00.000",
          "weather": "rain"
        },
        {
          "__count": 6,
          "__count_end": 38,
          "__count_start": 32,
          "month_date": "2012-12-01T00:00:00.000",
          "weather": "snow"
        },
        {
          "__count": 32,
          "__count_end": 32,
          "__count_start": 0,
          "month_date": "2012-12-01T00:00:00.000",
          "weather": "sun"
        }
      ],
      "transform": [
        {
          "type": "formula",
          "expr": "toDate(datum['month_date'])",
          "as": "month_date"
        }
      ]
    },
    {
      "name": "source_0_x_domain_month_date",
      "values": [
        {"month_date": "2012-01-01T00:00:00.000"},
        {"month_date": "2012-02-01T00:00:00.000"},
        {"month_date": "2012-03-01T00:00:00.000"},
        {"month_date": "2012-04-01T00:00:00.000"},
        {"month_date": "2012-05-01T00:00:00.000"},
        {"month_date": "2012-06-01T00:00:00.000"},
        {"month_date": "2012-07-01T00:00:00.000"},
        {"month_date": "2012-08-01T00:00:00.000"},
        {"month_date": "2012-09-01T00:00:00.000"},
        {"month_date": "2012-10-01T00:00:00.000"},
        {"month_date": "2012-11-01T00:00:00.000"},
        {"month_date": "2012-12-01T00:00:00.000"}
      ],
      "transform": [
        {
          "type": "formula",
          "expr": "toDate(datum['month_date'])",
          "as": "month_date"
        }
      ]
    }
  ],
  "signals": [
    {"name": "x_step", "value": 20},
    {
      "name": "width",
      "update": "bandspace(domain('x').length, 0.1, 0.05) * x_step"
    }
  ],
  "marks": [
    {
      "type": "rect",
      "name": "marks",
      "from": {"data": "source_0"},
      "encode": {
        "update": {
          "x": {"field": "month_date", "scale": "x"},
          "description": {
            "signal": "\"Month of the year: \" + (timeFormat(datum[\"month_date\"], timeUnitSpecifier([\"month\"], {\"year-month\":\"%b %Y \",\"year-month-date\":\"%b %d, %Y \"}))) + \"; Count of Records: \" + (format(datum[\"__count\"], \"\")) + \"; Weather type: \" + (isValid(datum[\"weather\"]) ? datum[\"weather\"] : \"\"+datum[\"weather\"])"
          },
          "fill": {"field": "weather", "scale": "color"},
          "ariaRoleDescription": {"value": "bar"},
          "width": {"signal": "max(0.25, bandwidth('x'))"},
          "y": {"field": "__count_end", "scale": "y"},
          "y2": {"field": "__count_start", "scale": "y"}
        }
      },
      "style": ["bar"]
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "band",
      "domain": {
        "data": "source_0_x_domain_month_date",
        "field": "month_date",
        "sort": true
      },
      "range": {"step": {"signal": "x_step"}},
      "paddingInner": 0.1,
      "paddingOuter": 0.05
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {
        "fields": ["__count_start", "__count_end"],
        "data": "source_0"
      },
      "range": [{"signal": "height"}, 0],
      "zero": true,
      "nice": true
    },
    {
      "name": "color",
      "type": "ordinal",
      "domain": ["sun", "fog", "drizzle", "rain", "snow"],
      "range": ["#e7ba52", "#c7c7c7", "#aec7e8", "#1f77b4", "#9467bd"]
    }
  ],
  "axes": [
    {
      "scale": "y",
      "domain": false,
      "minExtent": 0,
      "aria": false,
      "orient": "left",
      "tickCount": {"signal": "ceil(height/40)"},
      "grid": true,
      "labels": false,
      "maxExtent": 0,
      "ticks": false,
      "gridScale": "x",
      "zindex": 0
    },
    {
      "scale": "x",
      "formatType": "time",
      "grid": false,
      "orient": "bottom",
      "title": "Month of the year",
      "format": {
        "signal": "timeUnitSpecifier([\"month\"], {\"year-month\":\"%b %Y \",\"year-month-date\":\"%b %d, %Y \"})"
      },
      "labelOverlap": true,
      "zindex": 0
    },
    {
      "scale": "y",
      "labelOverlap": true,
      "orient": "left",
      "title": "Count of Records",
      "grid": false,
      "zindex": 0,
      "tickCount": {"signal": "ceil(height/40)"}
    }
  ],
  "padding": 5,
  "background": "white",
  "legends": [
    {"title": "Weather type", "fill": "color", "symbolType": "square"}
  ],
  "style": "cell",
  "height": 200
}
    """)



# https://altair-viz.github.io/gallery/top_k_with_others.html

import altair as alt
from vega_datasets import data

source = data.movies.url

alt.Chart(source).mark_bar().encode(
    x=alt.X("aggregate_gross:Q", aggregate="mean", title=None),
    y=alt.Y(
        "ranked_director:N",
        sort=alt.Sort(op="mean", field="aggregate_gross", order="descending"),
        title=None,
    ),
).transform_aggregate(
    aggregate_gross='mean(Worldwide_Gross)',
    groupby=["Director"],
).transform_window(
    rank='row_number()',
    sort=[alt.SortField("aggregate_gross", order="descending")],
).transform_calculate(
    ranked_director="datum.rank < 10 ? datum.Director : 'All Others'"
).properties(
    title="Top Directors by Average Worldwide Gross",
)



import altair as alt
from vega_datasets import data

source = data.seattle_weather()
brush = alt.selection_interval(encodings=['x'])

bars = alt.Chart().mark_bar().encode(
    x='month(date):O',
    y='mean(precipitation):Q',
    opacity=alt.condition(brush, alt.OpacityValue(1), alt.OpacityValue(0.7)),
).add_params(
    brush
)

line = alt.Chart().mark_rule(color='firebrick').encode(
    y='mean(precipitation):Q',
    size=alt.SizeValue(3)
).transform_filter(
    brush
)

alt.layer(bars, line, data=source)



import pandas as pd
import pytest
from pandas import Timestamp, NaT
import pyarrow as pa
import vegafusion as vf
import json
import polars as pl
from datetime import date
import decimal
import base64

def setup_module(module):
    vf.set_local_tz("UTC")


def get_connections():
    connections = ["datafusion"]
    try:
        import duckdb
        connections.append("duckdb")
    except ImportError:
        pass

    return connections


def order_items_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "height": 200,
  "style": "cell",
  "data": [
    {"name": "order_items", "url": "table://order_items"},
    {
      "name": "data_0",
      "source": "order_items",
      "transform": [
        {
          "type": "aggregate",
          "groupby": ["menu_item"],
          "ops": ["count"],
          "fields": ["menu_item"],
          "as": ["__count"]
        },
        {
            "type": "collect",
            "sort": {"field": ["menu_item"]}
        }
      ]
    }
  ],
  "signals": [
    {"name": "x_step", "value": 20},
    {
      "name": "width",
      "update": "bandspace(domain('x').length, 0.1, 0.05) * x_step"
    }
  ],
  "marks": [
    {
      "name": "layer_0_marks",
      "type": "rect",
      "clip": true,
      "style": ["bar"],
      "from": {"data": "data_0"},
      "encode": {
        "update": {
          "fill": {"value": "#4c78a8"},
          "x": {"scale": "x", "field": "menu_item"},
          "width": {"scale": "x", "band": 1},
          "y": {"scale": "y", "field": "__count"},
          "y2": {"scale": "y", "value": 0}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "band",
      "domain": {"data": "data_0", "field": "menu_item", "sort": true},
      "range": {"step": {"signal": "x_step"}},
      "paddingInner": 0.1,
      "paddingOuter": 0.05
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {"data": "data_0", "field": "__count"},
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    }
  ]
}
""")


def movies_histogram_spec(agg="count"):
    return json.loads("""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "width": 200,
  "height": 200,
  "style": "cell",
  "data": [
    {
      "name": "source_0",
      "url": "data/movies.json",
      "format": {"type": "json"},
      "transform": [
        {
          "type": "extent",
          "field": "IMDB Rating",
          "signal": "bin_maxbins_10_IMDB_Rating_extent"
        },
        {
          "type": "bin",
          "field": "IMDB Rating",
          "as": [
            "bin_maxbins_10_IMDB Rating",
            "bin_maxbins_10_IMDB Rating_end"
          ],
          "signal": "bin_maxbins_10_IMDB_Rating_bins",
          "extent": {"signal": "bin_maxbins_10_IMDB_Rating_extent"},
          "maxbins": 10
        },
        {
          "type": "aggregate",
          "groupby": [
            "bin_maxbins_10_IMDB Rating",
            "bin_maxbins_10_IMDB Rating_end"
          ],
          "ops": [""" + '"' + agg + '"' + r"""],
          "fields": ["Worldwide Gross"],
          "as": ["median_Worldwide Gross"]
        },
        {
          "type": "filter",
          "expr": "isValid(datum[\"bin_maxbins_10_IMDB Rating\"]) && isFinite(+datum[\"bin_maxbins_10_IMDB Rating\"]) && isValid(datum[\"median_Worldwide Gross\"]) && isFinite(+datum[\"median_Worldwide Gross\"])"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "marks",
      "type": "rect",
      "style": ["bar"],
      "from": {"data": "source_0"},
      "encode": {
        "update": {
          "fill": {"value": "#4c78a8"},
          "ariaRoleDescription": {"value": "bar"},
          "description": {
            "signal": "\"IMDB Rating (binned): \" + (!isValid(datum[\"bin_maxbins_10_IMDB Rating\"]) || !isFinite(+datum[\"bin_maxbins_10_IMDB Rating\"]) ? \"null\" : format(datum[\"bin_maxbins_10_IMDB Rating\"], \"\") + \" – \" + format(datum[\"bin_maxbins_10_IMDB Rating_end\"], \"\")) + \"; Median of Worldwide Gross: \" + (format(datum[\"median_Worldwide Gross\"], \"\"))"
          },
          "x2": {
            "scale": "x",
            "field": "bin_maxbins_10_IMDB Rating",
            "offset": 1
          },
          "x": {"scale": "x", "field": "bin_maxbins_10_IMDB Rating_end"},
          "y": {"scale": "y", "field": "median_Worldwide Gross"},
          "y2": {"scale": "y", "value": 0}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "linear",
      "domain": {
        "signal": "[bin_maxbins_10_IMDB_Rating_bins.start, bin_maxbins_10_IMDB_Rating_bins.stop]"
      },
      "range": [0, {"signal": "width"}],
      "bins": {"signal": "bin_maxbins_10_IMDB_Rating_bins"},
      "zero": false
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {"data": "source_0", "field": "median_Worldwide Gross"},
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    }
  ],
  "axes": [
    {
      "scale": "y",
      "orient": "left",
      "gridScale": "x",
      "grid": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "x",
      "orient": "bottom",
      "grid": false,
      "title": "IMDB Rating (binned)",
      "labelFlush": true,
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(width/10)"},
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": false,
      "title": "Median of Worldwide Gross",
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "zindex": 0
    }
  ]
}""")


def standalone_aggregate_spec(agg="count"):
    return json.loads("""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "data": [
    {
      "name": "df",
      "values": [
        {
          "x": "a",
          "y": 1
        },
        {
          "x": "a",
          "y": 2
        },
        {
          "x": "a",
          "y": 3
        }
      ]
    },
    {
      "name": "data_0",
      "source": "df",
      "transform": [
        {
          "type": "aggregate",
          "groupby": [
            "x"
          ],
          "fields": [
            "y"
          ],
          "ops": [
            """ + '"' + agg + '"' + """
          ],
          "as": [
            "median_y"
          ]
        },
        {
          "type": "filter",
          "expr": "isValid(datum['median_y']) && isFinite(+datum['median_y'])"
        },
        {
          "type": "project",
          "fields": [
            "median_y",
            "x"
          ]
        }
      ]
    },
    {
      "name": "data_0_x_domain_x",
      "source": "data_0",
      "transform": [
        {
          "type": "aggregate",
          "groupby": [
            "x"
          ],
          "fields": [],
          "ops": [],
          "as": []
        },
        {
          "type": "formula",
          "expr": "datum['x']",
          "as": "sort_field"
        }
      ]
    },
    {
      "name": "data_0_y_domain_median_y",
      "source": "data_0",
      "transform": [
        {
          "type": "formula",
          "expr": "+datum['median_y']",
          "as": "median_y"
        },
        {
          "type": "aggregate",
          "groupby": [],
          "fields": [
            "median_y",
            "median_y"
          ],
          "ops": [
            "min",
            "max"
          ],
          "as": [
            "min",
            "max"
          ]
        }
      ]
    }
  ],
  "signals": [
    {
      "name": "width",
      "init": "isFinite(containerSize()[0]) ? containerSize()[0] : 200",
      "on": [
        {
          "events": "window:resize",
          "update": "isFinite(containerSize()[0]) ? containerSize()[0] : 200"
        }
      ]
    },
    {
      "name": "height",
      "init": "isFinite(containerSize()[1]) ? containerSize()[1] : 200",
      "on": [
        {
          "events": "window:resize",
          "update": "isFinite(containerSize()[1]) ? containerSize()[1] : 200"
        }
      ]
    }
  ],
  "marks": [
    {
      "type": "rect",
      "name": "layer_0_marks",
      "from": {
        "data": "data_0"
      },
      "encode": {
        "update": {
          "tooltip": {
            "signal": "{'x': isValid(datum['x']) ? datum['x'] : ''+datum['x'], 'Median of y': format(datum['median_y'], '')}"
          },
          "fill": {
            "value": "#4c78a8"
          },
          "description": {
            "signal": "'x: ' + (isValid(datum['x']) ? datum['x'] : ''+datum['x']) + '; Median of y: ' + (format(datum['median_y'], ''))"
          },
          "width": {
            "scale": "x",
            "band": 1
          },
          "x": {
            "field": "x",
            "scale": "x"
          },
          "y": {
            "field": "median_y",
            "scale": "y"
          },
          "y2": {
            "value": 0,
            "scale": "y"
          },
          "ariaRoleDescription": {
            "value": "bar"
          }
        }
      },
      "style": [
        "bar"
      ],
      "clip": true
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "band",
      "domain": {
        "data": "data_0_x_domain_x",
        "field": "x",
        "sort": true
      },
      "range": [
        0,
        {
          "signal": "width"
        }
      ],
      "paddingOuter": 0.05,
      "paddingInner": 0.1
    },
    {
      "name": "y",
      "type": "linear",
      "domain": [
        {
          "signal": "(data('data_0_y_domain_median_y')[0] || {}).min"
        },
        {
          "signal": "(data('data_0_y_domain_median_y')[0] || {}).max"
        }
      ],
      "range": [
        {
          "signal": "height"
        },
        0
      ],
      "nice": true,
      "zero": true
    }
  ],
  "axes": [
    {
      "scale": "y",
      "grid": true,
      "minExtent": 0,
      "aria": false,
      "orient": "left",
      "gridScale": "x",
      "zindex": 0,
      "tickCount": {
        "signal": "ceil(height/40)"
      },
      "maxExtent": 0,
      "labels": false,
      "ticks": false,
      "domain": false
    },
    {
      "scale": "x",
      "orient": "bottom",
      "title": "x",
      "labelAlign": "right",
      "labelBaseline": "middle",
      "grid": false,
      "labelAngle": 270,
      "zindex": 0
    },
    {
      "scale": "y",
      "title": "Median of y",
      "zindex": 0,
      "labelOverlap": true,
      "tickCount": {
        "signal": "ceil(height/40)"
      },
      "orient": "left",
      "grid": false
    }
  ],
  "padding": 5,
  "background": "white",
  "autosize": {
    "type": "fit",
    "contains": "padding"
  },
  "config": {
    "customFormatTypes": true,
    "style": {
      "guide-label": {
        "font": "'IBM Plex Sans', system-ui, -apple-system, BlinkMacSystemFont, sans-serif"
      },
      "guide-title": {
        "font": "'IBM Plex Sans', system-ui, -apple-system, BlinkMacSystemFont, sans-serif"
      },
      "group-title": {
        "font": "'IBM Plex Sans', system-ui, -apple-system, BlinkMacSystemFont, sans-serif"
      },
      "group-subtitle": {
        "font": "'IBM Plex Sans', system-ui, -apple-system, BlinkMacSystemFont, sans-serif"
      },
      "cell": {},
      "text": {
        "font": "'IBM Plex Sans', system-ui, -apple-system, BlinkMacSystemFont, sans-serif"
      }
    }
  },
  "style": "cell",
  "usermeta": {
    "warnings": []
  }
}    
    """)


def date_column_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "description": "A scatterplot showing horsepower and miles per gallons for various cars.",
  "background": "white",
  "padding": 5,
  "width": 400,
  "height": 20,
  "style": "cell",
  "data": [
    {
      "name": "data_0",
      "url": "vegafusion+dataset://dates",
      "transform": [
        {
          "type": "formula",
          "expr": "toDate(datum[\"date_col\"])",
          "as": "date_col"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "marks",
      "type": "symbol",
      "style": ["point"],
      "from": {"data": "data_0"},
      "encode": {
        "update": {
          "opacity": {"value": 0.7},
          "fill": {"value": "transparent"},
          "stroke": {"value": "#4c78a8"},
          "ariaRoleDescription": {"value": "point"},
          "description": {
            "signal": "\"date32: \" + (timeFormat(datum[\"date32\"], '%b %d, %Y'))"
          },
          "x": {"scale": "x", "field": "date_col"},
          "y": {"signal": "height", "mult": 0.5}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "time",
      "domain": {"data": "data_0", "field": "date_col"},
      "range": [0, {"signal": "width"}]
    }
  ]
}
""")


def period_in_col_name_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "width": 200,
  "height": 200,
  "style": "cell",
  "data": [
    {
      "name": "df_period",
      "url": "vegafusion+dataset://df_period"
    },
    {
      "name": "data_0",
      "source": "df_period",
      "transform": [
        {
          "type": "filter",
          "expr": "isValid(datum[\"normal\"]) && isFinite(+datum[\"normal\"]) && isValid(datum[\"a.b\"]) && isFinite(+datum[\"a.b\"])"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "layer_0_marks",
      "type": "rect",
      "clip": true,
      "style": ["bar"],
      "from": {"data": "data_0"},
      "encode": {
        "update": {
          "tooltip": {
            "signal": "{\"normal\": format(datum[\"normal\"], \"\"), \"a\\.b\": format(datum[\"a.b\"], \"\")}"
          },
          "fill": {"value": "#4c78a8"},
          "ariaRoleDescription": {"value": "bar"},
          "description": {
            "signal": "\"normal: \" + (format(datum[\"normal\"], \"\")) + \"; a\\.b: \" + (format(datum[\"a.b\"], \"\"))"
          },
          "xc": {"scale": "x", "field": "normal"},
          "width": {"value": 5},
          "y": {"scale": "y", "field": "a\\.b"},
          "y2": {"scale": "y", "value": 0}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "linear",
      "domain": {"data": "data_0", "field": "normal"},
      "range": [0, {"signal": "width"}],
      "nice": true,
      "zero": false,
      "padding": 5
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {"data": "data_0", "field": "a\\.b"},
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    }
  ],
  "axes": [
    {
      "scale": "x",
      "orient": "bottom",
      "gridScale": "y",
      "grid": true,
      "tickCount": {"signal": "ceil(width/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "gridScale": "x",
      "grid": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "x",
      "orient": "bottom",
      "grid": false,
      "title": "normal",
      "labelFlush": true,
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(width/40)"},
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": false,
      "title": "a\\.b",
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "zindex": 0
    }
  ]
} 
    """)


def nat_bar_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "width": 200,
  "height": 200,
  "style": "cell",
  "data": [
    {
      "name": "dataframe",
      "url": "vegafusion+dataset://dataframe",
      "format": {"type": "json", "parse": {"NULL_TEST": "date"}},
      "transform": [
        {
          "type": "stack",
          "groupby": ["NULL_TEST"],
          "field": "SALES",
          "sort": {"field": [], "order": []},
          "as": ["SALES_start", "SALES_end"],
          "offset": "zero"
        },
        {
          "type": "filter",
          "expr": "(isDate(datum[\"NULL_TEST\"]) || (isValid(datum[\"NULL_TEST\"]) && isFinite(+datum[\"NULL_TEST\"]))) && isValid(datum[\"SALES\"]) && isFinite(+datum[\"SALES\"])"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "layer_0_marks",
      "type": "rect",
      "clip": true,
      "style": ["bar"],
      "from": {"data": "dataframe"},
      "encode": {
        "update": {
          "tooltip": {
            "signal": "{\"NULL_TEST\": timeFormat(datum[\"NULL_TEST\"], '%b %d, %Y'), \"SALES\": format(datum[\"SALES\"], \"\")}"
          },
          "fill": {"value": "#4c78a8"},
          "ariaRoleDescription": {"value": "bar"},
          "description": {
            "signal": "\"NULL_TEST: \" + (timeFormat(datum[\"NULL_TEST\"], '%b %d, %Y')) + \"; SALES: \" + (format(datum[\"SALES\"], \"\"))"
          },
          "xc": {"scale": "x", "field": "NULL_TEST"},
          "width": {"value": 5},
          "y": {"scale": "y", "field": "SALES_end"},
          "y2": {"scale": "y", "field": "SALES_start"}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "time",
      "domain": {"data": "dataframe", "field": "NULL_TEST"},
      "range": [0, {"signal": "width"}],
      "padding": 5
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {"data": "dataframe", "fields": ["SALES_start", "SALES_end"]},
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    }
  ],
  "axes": [
    {
      "scale": "x",
      "orient": "bottom",
      "gridScale": "y",
      "grid": true,
      "tickCount": {"signal": "ceil(width/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "gridScale": "x",
      "grid": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "x",
      "orient": "bottom",
      "grid": false,
      "title": "NULL_TEST",
      "labelFlush": true,
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(width/40)"},
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": false,
      "title": "SALES",
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "zindex": 0
    }
  ],
  "config": {
    "range": {"ramp": {"scheme": "yellowgreenblue"}},
    "axis": {"domain": false}
  }
}
    """)


def date32_timeunit_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "autosize": {
    "type": "fit",
    "contains": "padding"
  },
  "background": "white",
  "padding": 5,
  "style": "cell",
  "data": [
    {
      "name": "dataframe",
      "url": "vegafusion+dataset://dataframe"
    },
    {
      "name": "data_0",
      "source": "dataframe",
      "transform": [
        {
          "type": "filter",
          "expr": "isValid(datum[\"GO_LIVE_MONTH\"])"
        },
        {
          "field": "GO_LIVE_MONTH",
          "type": "timeunit",
          "units": [
            "year",
            "month"
          ],
          "as": [
            "go_live_start",
            "go_live_end"
          ]
        },
        {
          "type": "stack",
          "groupby": [
            "go_live_start"
          ],
          "field": "PERCENT_GO_LIVES",
          "sort": {
            "field": [],
            "order": []
          },
          "as": [
            "percent_go_lives_start",
            "percent_go_lives_end"
          ],
          "offset": "normalize"
        },
        {
          "type": "formula",
          "expr": "datum['percent_go_lives_end']-datum['percent_go_lives_start']",
          "as": "percent_go_lives_delta"
        }
      ]
    },
    {
      "name": "data_1",
      "source": "data_0",
      "transform": [
        {
          "field": "go_live_start",
          "type": "timeunit",
          "units": [
            "year",
            "month"
          ],
          "as": [
            "go_live_start_start",
            "go_live_start_end"
          ]
        },
        {
          "type": "filter",
          "expr": "(isDate(datum[\"go_live_start_start\"]) || (isValid(datum[\"go_live_start_start\"]) && isFinite(+datum[\"go_live_start_start\"]))) && isValid(datum[\"percent_go_lives_start\"]) && isFinite(+datum[\"percent_go_lives_start\"])"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "layer_0_layer_0_layer_0_marks",
      "type": "rect",
      "clip": true,
      "style": [
        "bar"
      ],
      "from": {
        "data": "data_1"
      },
      "encode": {
        "update": {
          "fill": {
            "value": "#ff4c00"
          },
          "opacity": {
            "value": 1
          },
          "ariaRoleDescription": {
            "value": "bar"
          },
          "x": {
            "scale": "x",
            "field": "go_live_start_start"
          },
          "x2": {
            "scale": "x",
            "field": "go_live_end",
            "offset": -1
          },
          "y": {
            "scale": "y",
            "field": "percent_go_lives_start"
          },
          "y2": {
            "scale": "y",
            "field": "percent_go_lives_end"
          }
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "time",
      "domain": {
        "fields": [
          {
            "data": "data_1",
            "field": "go_live_start_start"
          },
          {
            "data": "data_1",
            "field": "go_live_end"
          }
        ]
      },
      "range": [
        0,
        {
          "signal": "width"
        }
      ]
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {
        "fields": [
          {
            "data": "data_1",
            "field": "percent_go_lives_start"
          },
          {
            "data": "data_1",
            "field": "percent_go_lives_end"
          }
        ]
      },
      "range": [
        {
          "signal": "height"
        },
        0
      ],
      "nice": true,
      "zero": true
    }
  ]
}

""")


def gh_268_hang_spec():
    return json.loads(r""" 
    {
      "$schema": "https://vega.github.io/schema/vega/v5.json",
      "data": [
        {
          "name": "interval_intervalselection__store"
        },
        {
          "name": "legend_pointselection__store"
        },
        {
          "name": "pivot_hover_32f2e9aa_f08a_4fb5_aa8a_ab3f2cc94a1d_store"
        },
        {
          "name": "movies_clean",
          "url": "vegafusion+dataset://movies_clean"
        },
        {
          "name": "data_0",
          "source": "movies_clean",
          "transform": [
            {
              "type": "formula",
              "expr": "toDate(datum[\"Release Date\"])",
              "as": "Release Date"
            }
          ]
        },
        {
          "name": "data_2",
          "source": "data_0",
          "transform": [
            {
              "field": "Release Date",
              "type": "timeunit",
              "units": [
                "year"
              ],
              "as": [
                "year_Release Date",
                "year_Release Date_end"
              ]
            }
          ]
        },
        {
          "name": "data_3",
          "source": "data_2",
          "transform": [
            {
              "type": "filter",
              "expr": "!length(data(\"interval_intervalselection__store\")) || vlSelectionTest(\"interval_intervalselection__store\", datum)"
            },
            {
              "type": "filter",
              "expr": "time('1986-11-09T18:28:05.617') <= time(datum[\"year_Release Date\"]) && time(datum[\"year_Release Date\"]) <= time('2001-09-16T22:23:39.144')"
            },
            {
              "type": "filter",
              "expr": "!length(data(\"legend_pointselection__store\")) || vlSelectionTest(\"legend_pointselection__store\", datum)"
            }
          ]
        }
      ]
    }
    """)


def manual_histogram_spec():
    return json.loads(r"""
    {
      "$schema": "https://vega.github.io/schema/vega/v5.json",
      "background": "white",
      "padding": 5,
      "width": 200,
      "height": 200,
      "style": "cell",
      "data": [
        {
          "name": "movies",
          "url": "https://cdn.jsdelivr.net/npm/vega-datasets@v1.29.0/data/movies.json",
          "format": {"type": "json"},
          "transform": [
            {
              "type": "extent",
              "field": "IMDB_Rating",
              "signal": "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_extent"
            },
            {
              "type": "bin",
              "field": "IMDB_Rating",
              "as": ["__bin_field_name", "__bin_field_name_end"],
              "signal": "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_bins",
              "extent": {
                "signal": "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_extent"
              },
              "maxbins": 10
            },
            {
              "type": "aggregate",
              "groupby": ["__bin_field_name", "__bin_field_name_end"],
              "ops": ["count"],
              "fields": [null],
              "as": ["__count"]
            },
            {
              "type": "formula",
              "expr": "'[' + toString(datum[\"__bin_field_name\"]) + ', ' + toString(datum[\"__bin_field_name_end\"]) + ')'",
              "as": "__bin_range"
            },
            {
              "type": "filter",
              "expr": "isValid(datum[\"__bin_field_name\"]) && isFinite(+datum[\"__bin_field_name\"]) && isValid(datum[\"__count\"]) && isFinite(+datum[\"__count\"])"
            }
          ]
        }
      ],
      "marks": [
        {
          "name": "layer_0_layer_0_marks",
          "type": "rect",
          "clip": true,
          "style": ["bar"],
          "from": {"data": "movies"},
          "encode": {
            "update": {
              "fill": {"value": "#4C78A8"},
              "opacity": {"value": 1},
              "ariaRoleDescription": {"value": "bar"},
              "description": {
                "signal": "\"IMDB_Rating (start): \" + (format(datum[\"__bin_field_name\"], \"\")) + \"; Count of Records: \" + (format(datum[\"__count\"], \"\")) + \"; __bin_field_name_end: \" + (format(datum[\"__bin_field_name_end\"], \"\"))"
              },
              "x": {"scale": "x", "field": "__bin_field_name"},
              "x2": {"scale": "x", "field": "__bin_field_name_end", "offset": -1},
              "y": {"scale": "y", "field": "__count"},
              "y2": {"scale": "y", "value": 0}
            }
          }
        }
      ],
      "scales": [
        {
          "name": "x",
          "type": "linear",
          "domain": {
            "data": "movies",
            "fields": ["__bin_field_name", "__bin_field_name_end"]
          },
          "range": [0, {"signal": "width"}],
          "nice": true,
          "zero": true
        },
        {
          "name": "y",
          "type": "linear",
          "domain": {"fields": [{"data": "movies", "field": "__count"}, [0]]},
          "range": [{"signal": "height"}, 0],
          "nice": true,
          "zero": true
        }
      ],
      "axes": [
        {
          "scale": "x",
          "orient": "bottom",
          "grid": true,
          "tickCount": 10,
          "gridScale": "y",
          "domain": false,
          "labels": false,
          "aria": false,
          "maxExtent": 0,
          "minExtent": 0,
          "ticks": false,
          "zindex": 0
        },
        {
          "scale": "y",
          "orient": "left",
          "grid": true,
          "gridScale": "x",
          "tickCount": {"signal": "ceil(height/40)"},
          "domain": false,
          "labels": false,
          "aria": false,
          "maxExtent": 0,
          "minExtent": 0,
          "ticks": false,
          "zindex": 0
        },
        {
          "scale": "x",
          "orient": "bottom",
          "grid": false,
          "title": "IMDB_Rating (start)",
          "labelFlush": false,
          "labels": true,
          "tickCount": 10,
          "ticks": true,
          "labelOverlap": true,
          "zindex": 0
        },
        {
          "scale": "y",
          "orient": "left",
          "grid": false,
          "title": "Count of Records",
          "labelFlush": false,
          "labels": true,
          "ticks": true,
          "labelOverlap": true,
          "tickCount": {"signal": "ceil(height/40)"},
          "zindex": 0
        }
      ]
    }
    """)


def empty_histogram_spec():
    return json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "background": "white",
  "padding": 5,
  "width": 200,
  "height": 200,
  "style": "cell",
  "data": [
    {"name": "empty_df", "url": "table://empty_df"},
    {
      "name": "data_0",
      "source": "empty_df",
      "transform": [
        {
          "type": "extent",
          "field": "col",
          "signal": "layer_0_layer_0_bin_maxbins_10_col_extent"
        },
        {
          "type": "bin",
          "field": "col",
          "as": ["__bin_field_name", "__bin_field_name_end"],
          "signal": "layer_0_layer_0_bin_maxbins_10_col_bins",
          "extent": {"signal": "layer_0_layer_0_bin_maxbins_10_col_extent"},
          "maxbins": 10
        },
        {
          "type": "aggregate",
          "groupby": ["__bin_field_name", "__bin_field_name_end"],
          "ops": ["count"],
          "fields": [null],
          "as": ["__count"]
        },
        {
          "type": "formula",
          "expr": "'[' + toString(datum[\"__bin_field_name\"]) + ', ' + toString(datum[\"__bin_field_name_end\"]) + ')'",
          "as": "__bin_range"
        },
        {
          "type": "filter",
          "expr": "isValid(datum[\"__bin_field_name\"]) && isFinite(+datum[\"__bin_field_name\"]) && isValid(datum[\"__count\"]) && isFinite(+datum[\"__count\"])"
        }
      ]
    }
  ],
  "marks": [
    {
      "name": "layer_0_layer_0_marks",
      "type": "rect",
      "clip": true,
      "style": ["bar"],
      "from": {"data": "data_0"},
      "encode": {
        "update": {
          "cursor": {"value": "pointer"},
          "fill": {"value": "#3e277a"},
          "opacity": {"value": 1},
          "ariaRoleDescription": {"value": "bar"},
          "description": {
            "signal": "\"col (start): \" + (format(datum[\"__bin_field_name\"], \"\")) + \"; Count of Records: \" + (format(datum[\"__count\"], \"\")) + \"; __bin_field_name_end: \" + (format(datum[\"__bin_field_name_end\"], \"\"))"
          },
          "x": {"scale": "x", "field": "__bin_field_name"},
          "x2": {"scale": "x", "field": "__bin_field_name_end", "offset": -1},
          "y": {"scale": "y", "field": "__count"},
          "y2": {"scale": "y", "value": 0}
        }
      }
    }
  ],
  "scales": [
    {
      "name": "x",
      "type": "linear",
      "domain": {
        "data": "data_0",
        "fields": ["__bin_field_name", "__bin_field_name_end"]
      },
      "range": [0, {"signal": "width"}],
      "nice": true,
      "zero": true
    },
    {
      "name": "y",
      "type": "linear",
      "domain": {"fields": [{"data": "data_0", "field": "__count"}, [0]]},
      "range": [{"signal": "height"}, 0],
      "nice": true,
      "zero": true
    }
  ],
  "axes": [
    {
      "scale": "x",
      "orient": "bottom",
      "grid": true,
      "tickCount": 10,
      "gridScale": "y",
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": true,
      "gridScale": "x",
      "tickCount": {"signal": "ceil(height/40)"},
      "domain": false,
      "labels": false,
      "aria": false,
      "maxExtent": 0,
      "minExtent": 0,
      "ticks": false,
      "zindex": 0
    },
    {
      "scale": "x",
      "orient": "bottom",
      "grid": false,
      "title": "col (start)",
      "labelFlush": false,
      "labels": true,
      "tickCount": 10,
      "ticks": true,
      "labelOverlap": true,
      "zindex": 0
    },
    {
      "scale": "y",
      "orient": "left",
      "grid": false,
      "title": "Count of Records",
      "labelFlush": false,
      "labels": true,
      "ticks": true,
      "labelOverlap": true,
      "tickCount": {"signal": "ceil(height/40)"},
      "zindex": 0
    }
  ]
}
""")


def test_pre_transform_multi_partition():
    n = 4050
    order_items = pd.DataFrame({
        "menu_item": [0] * n + [1] * n
    })

    vega_spec = order_items_spec()
    new_spec, warnings = vf.runtime.pre_transform_spec(vega_spec, inline_datasets={
        "order_items": order_items,
    })

    assert new_spec["data"][1] == dict(
        name="data_0",
        values=[
            {"menu_item": 0, "__count": n},
            {"menu_item": 1, "__count": n},
        ]
    )


def test_pre_transform_cache_cleared():
    # Make sure that result changes when input DataFrame changes
    def check(n):
        order_items = pd.DataFrame({
            "menu_item": [0] * n + [1] * n
        })

        vega_spec = order_items_spec()
        new_spec, warnings = vf.runtime.pre_transform_spec(vega_spec, inline_datasets={
            "order_items": order_items,
        })

        assert new_spec["data"][1] == dict(
            name="data_0",
            values=[
                {"menu_item": 0, "__count": n},
                {"menu_item": 1, "__count": n},
            ]
        )

    check(16)
    check(32)


def test_pre_transform_datasets():
    n = 4050
    order_items = pd.DataFrame({
        "menu_item": [0] * n + [1] * (2 * n) + [2] * (3 * n)
    })

    vega_spec = order_items_spec()
    datasets, warnings = vf.runtime.pre_transform_datasets(
        vega_spec,
        ["data_0"],
        inline_datasets={
            "order_items": order_items,
        }
    )
    assert len(warnings) == 0
    assert len(datasets) == 1

    result = datasets[0]
    expected = pd.DataFrame({"menu_item": [0, 1, 2], "__count": [n, 2 * n, 3 * n]})
    pd.testing.assert_frame_equal(result, expected)


def test_pre_transform_planner_warning1():
    # Pre-transform with supported aggregate function should result in no warnings
    vega_spec = movies_histogram_spec("mean")
    datasets, warnings = vf.runtime.pre_transform_spec(vega_spec)
    assert len(warnings) == 0

    # Pre-transform with unsupported aggregate function should result in one warning
    vega_spec = movies_histogram_spec("ci0")
    datasets, warnings = vf.runtime.pre_transform_spec(vega_spec)
    assert len(warnings) == 1

    warning = warnings[0]
    assert warning["type"] == "Planner"
    assert "source_0" in warning["message"]


def test_pre_transform_planner_warning2():
    # Pre-transform with supported aggregate function should result in no warnings
    vega_spec = standalone_aggregate_spec("mean")
    datasets, warnings = vf.runtime.pre_transform_spec(vega_spec)
    assert len(warnings) == 0

    # Pre-transform with unsupported aggregate function should result in one warning
    vega_spec = standalone_aggregate_spec("ci0")
    datasets, warnings = vf.runtime.pre_transform_spec(vega_spec)
    assert len(warnings) == 1

    warning = warnings[0]
    assert warning["type"] == "Planner"
    assert "data_0" in warning["message"]


def test_date32_pre_transform_dataset():
    # Test to make sure that date32 columns are interpreted in the local timezone
    dates_df = pd.DataFrame({
        "date_col": [date(2022, 1, 1), date(2022, 1, 2), date(2022, 1, 3)],
    })
    spec = date_column_spec()

    (output_ds,), warnings = vf.runtime.pre_transform_datasets(
        spec, ["data_0"], "America/New_York", default_input_tz="UTC", inline_datasets=dict(dates=dates_df)
    )

    # Timestamps are in the local timezone, so they should be midnight local time
    assert list(output_ds.date_col) == [
        pd.Timestamp('2022-01-01 00:00:00-0500', tz='America/New_York'),
        pd.Timestamp('2022-01-02 00:00:00-0500', tz='America/New_York'),
        pd.Timestamp('2022-01-03 00:00:00-0500', tz='America/New_York')
    ]


def test_date32_in_timeunit_duckdb_crash():
    try:
        # Set this as the active connection
        vf.runtime.set_connection("duckdb")

        # order_items includes a table://order_items data url
        vega_spec = date32_timeunit_spec()
        dataframe = pd.DataFrame({
            "GO_LIVE_MONTH": [date(2021, 1, 1), date(2021, 2, 1)],
            "PERCENT_GO_LIVES": [0.2, 0.3],
        })

        datasets, warnings = vf.runtime.pre_transform_datasets(
            vega_spec,
            ["data_1"],
            inline_datasets=dict(dataframe=dataframe)
        )
        assert len(warnings) == 0
        assert len(datasets) == 1
        assert len(datasets[0]) == 2
    finally:
        vf.runtime.set_connection("datafusion")


def test_period_in_column_name():
    df_period = pd.DataFrame([[1, 2]], columns=['normal', 'a.b'])
    spec = period_in_col_name_spec()
    datasets, warnings = vf.runtime.pre_transform_datasets(spec, ["data_0"], inline_datasets=dict(
        df_period=df_period
    ))
    assert len(warnings) == 0
    assert len(datasets) == 1

    dataset = datasets[0]
    assert dataset.to_dict("records") == [{"normal": 1, "a.b": 2}]


def test_nat_values():
    dataframe = pd.DataFrame([
        {'ORDER_DATE': date(2011, 3, 1),
         'SALES': 457.568,
         'NULL_TEST': Timestamp('2011-03-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 3, 1),
         'SALES': 376.509,
         'NULL_TEST': Timestamp('2011-03-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 3, 1),
         'SALES': 362.25,
         'NULL_TEST': Timestamp('2011-03-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 3, 1),
         'SALES': 129.552,
         'NULL_TEST': Timestamp('2011-03-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 3, 1), 'SALES': 18.84, 'NULL_TEST': NaT},
        {'ORDER_DATE': date(2011, 4, 1),
         'SALES': 66.96,
         'NULL_TEST': Timestamp('2011-04-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 4, 1), 'SALES': 6.24, 'NULL_TEST': NaT},
        {'ORDER_DATE': date(2011, 6, 1),
         'SALES': 881.93,
         'NULL_TEST': Timestamp('2011-06-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 6, 1),
         'SALES': 166.72,
         'NULL_TEST': Timestamp('2011-06-01 00:00:00+0000', tz="UTC")},
        {'ORDER_DATE': date(2011, 6, 1), 'SALES': 25.92, 'NULL_TEST': NaT}
    ])

    spec = nat_bar_spec()

    datasets, warnings = vf.runtime.pre_transform_datasets(spec, ["dataframe"], inline_datasets=dict(
        dataframe=dataframe
    ))
    assert len(warnings) == 0
    assert len(datasets) == 1

    dataset = datasets[0]
    assert dataset.to_dict("records")[0] == {
        'NULL_TEST': pd.Timestamp('2011-03-01 00:00:00+0000', tz="UTC"),
        'ORDER_DATE': Timestamp('2011-03-01 00:00:00+0000', tz="UTC"),
        'SALES': 457.568,
        'SALES_end': 457.568,
        'SALES_start': 0.0,
    }


def test_pre_transform_dataset_dataframe_interface_protocol():
    try:
        import pyarrow.interchange
    except ImportError:
        pytest.skip("DataFrame interface protocol requires pyarrow 11.0.0 or later")

    n = 4050
    # Input a polars DataFrame (which follows the DataFrame Interface Protocol)
    order_items = pl.DataFrame({
        "menu_item": [0] * n + [1] * (2 * n) + [2] * (3 * n)
    })

    vega_spec = order_items_spec()
    datasets, warnings = vf.runtime.pre_transform_datasets(
        vega_spec,
        ["data_0"],
        inline_datasets={
            "order_items": order_items,
        }
    )
    assert len(warnings) == 0
    assert len(datasets) == 1

    result = datasets[0]

    # Result should be a polars DataFrame
    assert isinstance(result, pl.DataFrame)
    expected = pl.DataFrame({"menu_item": [0, 1, 2], "__count": [n, 2 * n, 3 * n]})
    assert result.frame_equal(expected)


def test_pre_transform_dataset_duckdb_conn():
    import duckdb

    n = 4050
    # Input a polars DataFrame (which follows the DataFrame Interface Protocol)
    order_items = pd.DataFrame({
        "menu_item": [0] * n + [1] * (2 * n) + [2] * (3 * n)
    })

    try:
        # Create duckdb connection and register order_items with duckdb
        conn = duckdb.connect()
        conn.register("order_items", order_items)

        # Set this as the active connection
        vf.runtime.set_connection(conn)

        # order_items includes a table://order_items data url
        vega_spec = order_items_spec()
        datasets, warnings = vf.runtime.pre_transform_datasets(
            vega_spec,
            ["data_0"],
        )
        assert len(warnings) == 0
        assert len(datasets) == 1

        result = datasets[0]
        expected = pd.DataFrame({"menu_item": [0, 1, 2], "__count": [n, 2 * n, 3 * n]})
        pd.testing.assert_frame_equal(result, expected)
    finally:
        vf.runtime.set_connection("datafusion")


def test_pre_transform_dataset_duckdb_with_decimal_conn():
    import duckdb

    n = 4050
    # Input a polars DataFrame (which follows the DataFrame Interface Protocol)
    order_items = pd.DataFrame({
        "menu_item_int": [0] * n + [1] * (2 * n) + [2] * (3 * n)
    })

    try:
        # Create duckdb connection and register order_items with duckdb
        conn = duckdb.connect()
        conn.register("order_items_int", order_items)
        conn.query(
            "SELECT menu_item_int::DECIMAL(12,2) as menu_item from order_items_int"
        ).to_view("order_items")

        # Set this as the active connection
        vf.runtime.set_connection(conn)

        # order_items includes a table://order_items data url
        vega_spec = order_items_spec()
        datasets, warnings = vf.runtime.pre_transform_datasets(
            vega_spec,
            ["data_0"],
        )
        assert len(warnings) == 0
        assert len(datasets) == 1

        result = datasets[0]
        expected = pd.DataFrame({
            "menu_item": [decimal.Decimal(0), decimal.Decimal(1), decimal.Decimal(2)],
            "__count": [n, 2 * n, 3 * n]
        })
        pd.testing.assert_frame_equal(result, expected)
    finally:
        vf.runtime.set_connection("datafusion")


def test_duckdb_timestamp_with_timezone():
    try:
        vf.runtime.set_connection("duckdb")
        dates_df = pd.DataFrame({
            "date_col": [date(2022, 1, 1), date(2022, 1, 2), date(2022, 1, 3)],
        })
        dates_df["date_col"] = pd.to_datetime(dates_df.date_col).dt.tz_localize("UTC")
        spec = date_column_spec()

        (output_ds,), warnings = vf.runtime.pre_transform_datasets(
            spec, ["data_0"], "America/New_York", default_input_tz="UTC", inline_datasets=dict(dates=dates_df)
        )

        # Timestamps are in the local timezone, so they should be midnight local time
        assert list(output_ds.date_col) == [
            pd.Timestamp('2022-01-01 00:00:00', tz='UTC'),
            pd.Timestamp('2022-01-02 00:00:00', tz='UTC'),
            pd.Timestamp('2022-01-03 00:00:00', tz='UTC')
        ]
    finally:
        vf.runtime.set_connection("datafusion")


def test_gh_268_hang():
    """
    Tests for hang reported in https://github.com/hex-inc/vegafusion/issues/268
    """
    vf.runtime.set_connection("datafusion")
    movies = pd.read_json("https://raw.githubusercontent.com/vega/vega-datasets/main/data/movies.json")
    spec = gh_268_hang_spec()
    for i in range(20):
        # Break cache by removing one row each iteration
        movies_inner = movies.iloc[i:]
        vf.runtime.pre_transform_datasets(spec, ["data_3"], inline_datasets=dict(movies_clean=movies_inner))


def test_repeat_duckdb():
    """
    Tests for hang reported in https://github.com/hex-inc/vegafusion/issues/268
    """
    vf.runtime.set_connection("duckdb")
    movies = pd.read_json("https://raw.githubusercontent.com/vega/vega-datasets/main/data/movies.json")
    spec = gh_268_hang_spec()
    for i in range(2):
        vf.runtime.pre_transform_datasets(spec, ["data_3"], inline_datasets=dict(movies_clean=movies))


@pytest.mark.parametrize("connection", get_connections())
def test_pivot_mixed_case(connection):
    vf.runtime.set_connection(connection)
    source_0 = pd.DataFrame.from_records([
        {"country": "Norway", "type": "gold", "count": 14},
        {"country": "Norway", "type": "silver", "count": 14},
        {"country": "Norway", "type": "Gold", "count": 11},
        {"country": "Germany", "type": "gold", "count": 14},
        {"country": "Germany", "type": "silver", "count": 10},
        {"country": "Germany", "type": "bronze", "count": 7},
        {"country": "Canada", "type": "gold", "count": 11},
        {"country": "Canada", "type": "silver", "count": 8},
        {"country": "Canada", "type": "bronze", "count": 10}
    ])
    spec = json.loads(r"""
{
  "$schema": "https://vega.github.io/schema/vega/v5.json",
  "data": [
    {
      "name": "source_0",
      "url": "table://source_0"
    },
    {
      "name": "data_0",
      "source": "source_0",
      "transform": [
        {
          "type": "pivot",
          "field": "type",
          "value": "count",
          "groupby": ["country"]
        }
      ]
    }
  ]
}   
    """)

    datasets, warnings = vf.runtime.pre_transform_datasets(
        spec, ["data_0"], inline_datasets=dict(source_0=source_0)
    )

    assert set(datasets[0].columns.tolist()) == {"gold", "Gold", "silver", "bronze", "country"}


def test_keep_signals():
    spec = manual_histogram_spec()

    # pre-transform without keep_signals. No signals should be present in pre-transformed spec
    tx_spec, warnings = vf.runtime.pre_transform_spec(spec)
    assert len(tx_spec.get("signals", [])) == 0

    # Specify single keep_signal as a string
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        spec,
        keep_signals="layer_0_layer_0_bin_maxbins_10_IMDB_Rating_bins"
    )
    assert len(tx_spec.get("signals", [])) == 1
    sig0 = tx_spec["signals"][0]
    assert sig0["name"] == "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_bins"
    assert sig0["value"]["step"] == 1.0

    # Specify multiple keep_signals as a list
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        spec,
        keep_signals=[
            "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_bins",
            ("layer_0_layer_0_bin_maxbins_10_IMDB_Rating_extent", [])
        ]
    )
    assert len(tx_spec.get("signals", [])) == 2
    sig0 = tx_spec["signals"][0]
    sig1 = tx_spec["signals"][1]
    [sig0, sig1] = sorted([sig0, sig1], key=lambda v: v["name"])
    assert sig0["name"] == "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_bins"
    assert sig1["name"] == "layer_0_layer_0_bin_maxbins_10_IMDB_Rating_extent"
    assert sig1["value"] == [1.4, 9.2]


def test_empty_histogram():
    spec = empty_histogram_spec()
    empty_df = pd.DataFrame({ 'col': []})
    (data_0,), warnings = vf.runtime.pre_transform_datasets(
        spec, ["data_0"], inline_datasets=dict(empty_df=empty_df)
    )
    assert data_0.empty
    assert data_0.columns.tolist() == ["__bin_field_name", "__bin_field_name_end", "__count", "__bin_range"]


def test_pre_transform_spec_encoded_datasets():
    # Pre-transform with supported aggregate function should result in no warnings
    vega_spec = movies_histogram_spec()

    # default list of dict format
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        vega_spec, data_encoding_threshold=10, data_encoding_format="pyarrow"
    )

    values = tx_spec["data"][0]["values"]
    assert isinstance(values, list)
    assert len(values) == 9

    # pyarrow format
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        vega_spec, data_encoding_threshold=0, data_encoding_format="pyarrow"
    )

    values = tx_spec["data"][0]["values"]
    assert isinstance(values, pa.Table)
    values_df = values.to_pandas()
    assert len(values_df) == 9
    assert values_df.columns[0] == "bin_maxbins_10_IMDB Rating"

    # arrow-ipc format
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        vega_spec, data_encoding_threshold=0, data_encoding_format="arrow-ipc"
    )

    values = tx_spec["data"][0]["values"]
    assert isinstance(values, bytes)
    values_df = pa.ipc.deserialize_pandas(values)
    assert len(values_df) == 9
    assert values_df.columns[0] == "bin_maxbins_10_IMDB Rating"

    # arrow-ipc-base64 format
    tx_spec, warnings = vf.runtime.pre_transform_spec(
        vega_spec, data_encoding_threshold=0, data_encoding_format="arrow-ipc-base64"
    )

    values = tx_spec["data"][0]["values"]
    assert isinstance(values, str)
    values_df = pa.ipc.deserialize_pandas(base64.standard_b64decode(values))
    assert len(values_df) == 9
    assert values_df.columns[0] == "bin_maxbins_10_IMDB Rating"




# https://altair-viz.github.io/gallery/us_population_over_time_facet.html

import altair as alt
from vega_datasets import data

source = data.population.url

alt.Chart(source).mark_area().encode(
    x='age:O',
    y=alt.Y(
        'sum(people):Q',
        title='Population',
        axis=alt.Axis(format='~s')
    ),
    facet=alt.Facet('year:O', columns=5),
).properties(
    title='US Age Distribution By Year',
    width=80,
    height=80
)



# https://altair-viz.github.io/gallery/poly_fit_regression.html

import numpy as np
import pandas as pd
import altair as alt

# Generate some random data
np.random.seed(42)
rng = np.random.RandomState(1)
x = rng.rand(40) ** 2
y = 10 - 1.0 / (x + 0.1) + rng.randn(40)
source = pd.DataFrame({"x": x, "y": y})

# Define the degree of the polynomial fits
degree_list = [1, 3, 5]

base = alt.Chart(source).mark_circle(color="black").encode(
    alt.X("x"), alt.Y("y")
)

polynomial_fit = [
    base.transform_regression(
        "x", "y", method="poly", order=order, as_=["x", str(order)]
    )
        .mark_line()
        .transform_fold([str(order)], as_=["degree", "y"])
        .encode(alt.Color("degree:N"))
    for order in degree_list
]

alt.layer(base, *polynomial_fit)



import altair as alt
from vega_datasets import data
import vegafusion as vf
from io import StringIO, BytesIO
import json


def make_histogram():
    source = data.movies.url
    return alt.Chart(source).mark_bar().encode(
        alt.X("IMDB_Rating:Q", bin=True),
        y='count()',
    )


def test_save_html():
    chart = make_histogram()
    f = StringIO()
    vf.save_html(chart, f)
    html = f.getvalue().strip()
    assert html.startswith("<!DOCTYPE html>")
    assert "https://cdn.jsdelivr.net/npm/vega@5" in html
    assert "https://cdn.jsdelivr.net/npm/vega-embed@6" in html


# We can't test inline HTML until after a new release of Altair Viewer
# def test_save_html_inline():
#     chart = make_histogram()
#     f = StringIO()
#     vf.save_html(chart, f, inline=True)
#     html = f.getvalue().strip()
#     assert html.startswith("<!DOCTYPE html>")
#     assert "<script type=\"text/javascript\">" in html
#     assert "https://cdn.jsdelivr.net/npm/vega@5" not in html
#     assert "https://cdn.jsdelivr.net/npm/vega-embed@6" not in html


def test_save_vega():
    chart = make_histogram()
    f = StringIO()
    vf.save_vega(chart, f)
    vega = json.loads(f.getvalue())
    assert vega["$schema"] == "https://vega.github.io/schema/vega/v5.json"


def test_save_svg():
    chart = make_histogram()
    f = StringIO()
    vf.save_svg(chart, f)
    svg = f.getvalue()
    assert svg.startswith("<svg")


def test_save_png():
    chart = make_histogram()
    f = BytesIO()
    vf.save_png(chart, f)
    png = f.getvalue()
    assert png.startswith(b'\x89PNG')



import io
import os
import pathlib
from hashlib import sha1
from tempfile import NamedTemporaryFile
import uuid
import altair as alt
import pandas as pd
import pyarrow as pa
from weakref import WeakValueDictionary

DATASET_PREFIXES = ("vegafusion+dataset://", "table://")
BATCH_SIZE = 8096


def to_arrow_table(data):
    """
    Helper to convert a Pandas DataFrame to a PyArrow Table

    :param data: Pandas DataFrame
    :return: pyarrow.Table
    """
    import pyarrow as pa

    # Reset named index(ex) into a column
    if getattr(data.index, "name", None) is not None:
        data = data.reset_index()

    # Use pyarrow to infer schema from DataFrame
    for col, pd_type in data.dtypes.items():
        try:
            candidate_schema = pa.Schema.from_pandas(data[[col]])

            # Get inferred pyarrow type
            pa_type = candidate_schema.field(col).type

            # Convert Decimal columns to float
            if isinstance(pa_type, (pa.Decimal128Type, pa.Decimal256Type)):
                data = data.assign(**{col: data[col].astype("float64")})

        except pa.ArrowTypeError:
            if pd_type.kind == "O":
                # Try converting object columns to strings to handle cases where a
                # column has a mix of numbers and strings
                data = data.assign(**{col: data[col].astype(str)})

        # Expand categoricals (not yet supported in VegaFusion)
        if isinstance(pd_type, pd.CategoricalDtype):
            cat = data[col].cat
            data = data.assign(**{col: cat.categories[cat.codes]})

        # Copy un-aligned columns to align them
        # (arrow-rs seems to have trouble with un-aligned arrays)
        values = getattr(data[col], "values", None)
        if values is not None:
            flags = getattr(values, "flags", None)
            if flags is not None:
                try:
                    aligned = flags["ALIGNED"]
                    if not aligned:
                        data = data.assign(**{col: data[col].copy()})
                except IndexError:
                    pass

    # Convert DataFrame to table. Keep index only if named
    preserve_index = bool([name for name in getattr(data.index, "names", []) if name])
    table = pa.Table.from_pandas(data, preserve_index=preserve_index)

    # Split into batches of desired size
    batches = table.to_batches(BATCH_SIZE)
    table = pa.Table.from_batches(batches, table.schema)

    return table


def to_arrow_ipc_bytes(data, stream=False):
    """
    Helper to convert a DataFrame to the Arrow IPC binary format

    :param data: Pandas DataFrame, pyarrow Table, or object that supports
        the DataFrame Interchange Protocol
    :param stream: If True, write IPC Stream format. If False (default), write ipc file format.
    :return: bytes
    """
    if isinstance(data, pd.DataFrame):
        table = to_arrow_table(data)
    elif isinstance(data, pa.Table):
        table = data
    elif hasattr(data, "__dataframe__"):
        pi = import_pyarrow_interchange()
        table = pi.from_dataframe(data)
    else:
        raise ValueError(f"Unsupported input to to_arrow_ipc_bytes: {type(data)}")
    return arrow_table_to_ipc_bytes(table, stream=stream)


def arrow_table_to_ipc_bytes(table, stream=False):
    import pyarrow as pa

    # Next we write the Arrow table as a feather file (The Arrow IPC format on disk).
    # Write it in memory first so we can hash the contents before touching disk.
    bytes_buffer = io.BytesIO()

    if stream:
        with pa.ipc.new_stream(bytes_buffer, table.schema) as f:
            f.write_table(table, max_chunksize=BATCH_SIZE)
    else:
        with pa.ipc.new_file(bytes_buffer, table.schema) as f:
            f.write_table(table, max_chunksize=BATCH_SIZE)

    return bytes_buffer.getvalue()


def to_feather(data, file):
    """
    Helper to convert a Pandas DataFrame to a feather file that is optimized for
    use as a VegaFusion data source

    :param data: Pandas DataFrame, pyarrow Table, or object that supports
        the DataFrame Interchange Protocol
    :param file: File path string or writable file-like object
    :return: None
    """
    file_bytes = to_arrow_ipc_bytes(data, stream=False)

    # Either write to new file at path, or to writable file-like object
    if hasattr(file, "write"):
        file.write(file_bytes)
    else:
        with open(file, "wb") as f:
            f.write(file_bytes)


def feather_transformer(data, data_dir="_vegafusion_data"):
    from vegafusion import runtime

    if "vegafusion" not in alt.renderers.active:
        # Use default transformer if a vegafusion renderer is not active
        return alt.default_data_transformer(data)
    elif has_geo_interface(data):
        # Use default transformer for geo interface objects
        # (e.g. a geopandas GeoDataFrame)
        return alt.default_data_transformer(data)
    elif is_dataframe_like(data):
        if runtime.using_grpc:
            raise ValueError(
                "DataFrame data sources are not yet supported by the gRPC runtime"
            )

        bytes_buffer = io.BytesIO()
        to_feather(data, bytes_buffer)
        file_bytes = bytes_buffer.getvalue()

        # Hash bytes to generate unique file name
        hasher = sha1()
        hasher.update(file_bytes)
        hashstr = hasher.hexdigest()
        fname = f"vegafusion-{hashstr}.feather"

        # Check if file already exists
        tmp_dir = pathlib.Path(data_dir) / "tmp"
        os.makedirs(tmp_dir, exist_ok=True)
        path = pathlib.Path(data_dir) / fname
        if not path.is_file():
            # Write to temporary file then move (os.replace) to final destination. This is more resistant
            # to race conditions
            with NamedTemporaryFile(dir=tmp_dir, delete=False) as tmp_file:
                tmp_file.write(file_bytes)
                tmp_name = tmp_file.name

            os.replace(tmp_name, path)

        return {"url": path.as_posix()}
    else:
        # Use default transformer if we don't recognize data
        return alt.default_data_transformer(data)


def get_inline_dataset_names(vega_spec):
    """
    Get set of the inline datasets names in the provided spec

    :param vega_spec: Vega spec
    :return: set of inline dataset names
    """
    table_names = set()
    for data in vega_spec.get("data", []):
        url = data.get("url", "")
        for prefix in DATASET_PREFIXES:
            if url.startswith(prefix):
                name = url[len(prefix):]
                table_names.add(name)

    for mark in vega_spec.get("marks", []):
        table_names.update(get_inline_dataset_names(mark))

    return table_names


__inline_tables = WeakValueDictionary()


def get_inline_dataset_table(table_name):
    return __inline_tables.pop(table_name)


def get_inline_datasets_for_spec(vega_spec):
    table_names = get_inline_dataset_names(vega_spec)
    datasets = {}
    for table_name in table_names:
        try:
            datasets[table_name] = get_inline_dataset_table(table_name)
        except KeyError:
            # named dataset that was provided by the user
            pass
    return datasets


def inline_data_transformer(data):
    if has_geo_interface(data):
        # Use default transformer for geo interface objects
        # # (e.g. a geopandas GeoDataFrame)
        return alt.default_data_transformer(data)
    elif is_dataframe_like(data):
        table_name = f"table_{uuid.uuid4()}".replace("-", "_")
        __inline_tables[table_name] = data
        return {"url": DATASET_PREFIXES[0] + table_name}
    else:
        # Use default transformer if we don't recognize data
        return alt.default_data_transformer(data)


def is_dataframe_like(data):
    return isinstance(data, (pd.DataFrame, pa.Table)) or hasattr(data, "__dataframe__")


def has_geo_interface(data):
    return hasattr(data, "__geo_interface__")


def import_pyarrow_interchange():
    try:
        import pyarrow.interchange as pi
        return pi
    except ImportError:
        raise ImportError(
            "Use of the DataFrame Interchange Protocol requires at least version 11.0.0 of pyarrow\n"
            f"Found version {pa.__version__}"
        )


alt.data_transformers.register("vegafusion-feather", feather_transformer)
alt.data_transformers.register("vegafusion-inline", inline_data_transformer)
